# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the data classes for the `transaction settlement` ABCI application."""

import textwrap
from abc import ABC
from collections import deque
from enum import Enum
from typing import Any, Deque, Dict, List, Optional, Set, Tuple, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
    DegenerateRound,
    NONE_EVENT_ATTRIBUTE,
    OnlyKeeperSendsRound,
    get_name,
)
from packages.valory.skills.squads_transaction_settlement_abci.payloads import (
    ApproveTxPayload,
    CreateTxPayload,
    ExecuteTxPayload,
    RandomnessPayload,
    SelectKeeperPayload,
    VerifyTxPayload,
)


ADDRESS_LENGTH = 42
TX_HASH_LENGTH = 66
RETRIES_LENGTH = 64


class Event(Enum):
    """Event enumeration for the price estimation demo."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def is_keeper_set(self) -> bool:
        """Check whether keeper is set."""
        return bool(self.db.get("keepers", False))

    @property
    def keepers(self) -> Deque[str]:
        """Get the current cycle's keepers who have tried to submit a transaction."""
        if self.is_keeper_set:
            keepers_unparsed = cast(str, self.db.get_strict("keepers"))
            keepers_parsed = textwrap.wrap(
                keepers_unparsed[RETRIES_LENGTH:], ADDRESS_LENGTH
            )
            return deque(keepers_parsed)
        return deque()

    @property
    def most_voted_instruction_set(self) -> List[Dict[str, Any]]:
        """Return the list of instructions selected for execution on the multisig."""
        return cast(
            List[Dict[str, Any]], self.db.get_strict("most_voted_instruction_set")
        )

    @property
    def tx_pda(self) -> str:
        """Get the verified tx hash."""
        return cast(str, self.db.get_strict("tx_pda"))

    @property
    def most_voted_randomness_round(self) -> int:  # pragma: no cover
        """Get the first in priority keeper to try to re-submit a transaction."""
        round_ = self.db.get_strict("most_voted_randomness_round")
        return cast(int, round_)

    @property
    def most_voted_keeper_address(self) -> str:
        """Get the first in priority keeper to try to re-submit a transaction."""
        return self.keepers[0]


class CreateTxRandomnessRound(CollectSameUntilThresholdRound):
    """A round for generating randomness"""

    payload_class = RandomnessPayload
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = get_name(SynchronizedData.participant_to_randomness)
    selection_key = (
        get_name(SynchronizedData.most_voted_randomness_round),
        get_name(SynchronizedData.most_voted_randomness),
    )
    # the none event is not required because the `RandomnessPayload` does not allow for `None` values
    extended_requirements = tuple(
        attribute
        for attribute in CollectSameUntilThresholdRound.extended_requirements
        if attribute != NONE_EVENT_ATTRIBUTE
    )


class CreateTxSelectKeeperRound(CollectSameUntilThresholdRound):
    """A round in which a keeper is selected for transaction submission"""

    payload_class = SelectKeeperPayload
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = get_name(SynchronizedData.participant_to_selection)
    selection_key = get_name(SynchronizedData.keepers)
    # the none event is not required because the `SelectKeeperPayload` does not allow for `None` values
    extended_requirements = tuple(
        attribute
        for attribute in CollectSameUntilThresholdRound.extended_requirements
        if attribute != NONE_EVENT_ATTRIBUTE
    )


class CreateTxRound(OnlyKeeperSendsRound):
    """A round in which a keeper is selected for transaction submission"""

    keeper_payload: Optional[CreateTxPayload] = None
    payload_class = CreateTxPayload
    synchronized_data_class = SynchronizedData
    extended_requirements = ()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Process the end of the block."""
        if self.keeper_payload is None:
            return None
        synchronized_data = cast(
            SynchronizedData,
            self.synchronized_data.update(
                synchronized_data_class=self.synchronized_data_class,
                **{
                    get_name(SynchronizedData.tx_pda): self.keeper_payload.tx_pda,
                },
            ),
        )
        return synchronized_data, Event.DONE


class ApproveTxRound(CollectSameUntilThresholdRound):
    """Approve transaction round."""

    payload_class = ApproveTxPayload
    synchronized_data_class = SynchronizedData
    extended_requirements = ()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """End block."""
        if self.threshold_reached:
            return self.synchronized_data, Event.DONE
        return None


class ExecuteTxRandomnessRound(CreateTxRandomnessRound):
    """A round for generating randomness"""


class ExecuteTxSelectKeeperRound(CreateTxSelectKeeperRound):
    """A round in which a keeper is selected for transaction submission"""


class ExecuteTxRound(OnlyKeeperSendsRound):
    """Execute tx round."""

    keeper_payload: Optional[ExecuteTxPayload] = None
    payload_class = CreateTxPayload
    synchronized_data_class = SynchronizedData
    extended_requirements = ()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """End block."""
        if self.keeper_payload is None:
            return None
        synchronized_data = cast(
            SynchronizedData,
            self.synchronized_data.update(
                synchronized_data_class=self.synchronized_data_class,
                **{
                    get_name(SynchronizedData.tx_pda): self.keeper_payload.tx_pda,
                },
            ),
        )
        return synchronized_data, Event.DONE


class VerifyTxRound(CollectSameUntilThresholdRound):
    """Verify transaction round."""

    payload_class = VerifyTxPayload
    synchronized_data_class = SynchronizedData
    extended_requirements = ()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """End block."""
        # TODO: Fix implementation
        # This round makes an assumption that the transaction was executed successfully
        # and the value for `VerifyTxPayload.verified` is `True`.
        # This is to keep things simple for now.
        # We'll continue with a proper round implementation in the next iteration.
        if self.threshold_reached:
            return self.synchronized_data, Event.DONE
        return None


class FailedRound(DegenerateRound, ABC):
    """A round that represents that the period failed"""


class FinishedTransactionSubmissionRound(DegenerateRound, ABC):
    """A round that represents the transition to the ResetAndPauseRound"""


class SolanaTransactionSubmissionAbciApp(AbciApp[Event]):
    """SolanaTransactionSubmissionAbciApp

    Initial round: CreateTxRandomnessRound

    Initial states: {CreateTxRandomnessRound}

    Transition states:
        0. CreateTxRandomnessRound
            - done: 1.
            - no majority: 0.
            - round timeout: 0.
        1. CreateTxSelectKeeperRound
            - done: 2.
            - no majority: 0.
            - round timeout: 1.
        2. CreateTxRound
            - done: 3.
            - round timeout: 0.
        3. ApproveTxRound
            - done: 4.
            - no majority: 3.
            - round timeout: 3.
        4. ExecuteTxRandomnessRound
            - done: 5.
            - no majority: 4.
            - round timeout: 4.
        5. ExecuteTxSelectKeeperRound
            - done: 6.
            - no majority: 0.
            - round timeout: 5.
        6. ExecuteTxRound
            - done: 7.
            - round timeout: 4.
        7. VerifyTxRound
            - done: 8.
            - no majority: 7.
            - round timeout: 7.
        8. FinishedTransactionSubmissionRound
        9. FailedRound

    Final states: {FailedRound, FinishedTransactionSubmissionRound}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls: AppState = CreateTxRandomnessRound
    initial_states: Set[AppState] = {CreateTxRandomnessRound}
    transition_function: AbciAppTransitionFunction = {
        CreateTxRandomnessRound: {
            Event.DONE: CreateTxSelectKeeperRound,
            Event.NO_MAJORITY: CreateTxRandomnessRound,
            Event.ROUND_TIMEOUT: CreateTxRandomnessRound,
        },
        CreateTxSelectKeeperRound: {
            Event.DONE: CreateTxRound,
            Event.NO_MAJORITY: CreateTxRandomnessRound,
            Event.ROUND_TIMEOUT: CreateTxSelectKeeperRound,
        },
        CreateTxRound: {
            Event.DONE: ApproveTxRound,
            Event.ROUND_TIMEOUT: CreateTxRandomnessRound,
        },
        ApproveTxRound: {
            Event.DONE: ExecuteTxRandomnessRound,
            Event.NO_MAJORITY: ApproveTxRound,
            Event.ROUND_TIMEOUT: ApproveTxRound,
        },
        ExecuteTxRandomnessRound: {
            Event.DONE: ExecuteTxSelectKeeperRound,
            Event.NO_MAJORITY: ExecuteTxRandomnessRound,
            Event.ROUND_TIMEOUT: ExecuteTxRandomnessRound,
        },
        ExecuteTxSelectKeeperRound: {
            Event.DONE: ExecuteTxRound,
            Event.NO_MAJORITY: CreateTxRandomnessRound,
            Event.ROUND_TIMEOUT: ExecuteTxSelectKeeperRound,
        },
        ExecuteTxRound: {
            Event.DONE: VerifyTxRound,
            Event.ROUND_TIMEOUT: ExecuteTxRandomnessRound,
        },
        VerifyTxRound: {
            Event.DONE: FinishedTransactionSubmissionRound,
            Event.NO_MAJORITY: VerifyTxRound,
            Event.ROUND_TIMEOUT: VerifyTxRound,
        },
        FinishedTransactionSubmissionRound: {},
        FailedRound: {},
    }
    final_states: Set[AppState] = {
        FinishedTransactionSubmissionRound,
        FailedRound,
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
    }
    db_pre_conditions: Dict[AppState, Set[str]] = {
        CreateTxRandomnessRound: {
            get_name(SynchronizedData.most_voted_instruction_set),
            get_name(SynchronizedData.participants),
        }
    }
    db_post_conditions: Dict[AppState, Set[str]] = {
        FinishedTransactionSubmissionRound: {
            get_name(SynchronizedData.tx_pda),
        },
        FailedRound: set(),
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Image helpers."""

from pathlib import Path
from typing import Optional, Tuple

import click
from aea.configurations.data_types import Dependency, PublicId

from autonomy.configurations.loader import load_service_config
from autonomy.deploy.image import build_image as _build_image


def build_image(  # pylint: disable=too-many-arguments
    agent: Optional[PublicId],
    service_dir: Optional[Path],
    pull: bool = False,
    version: Optional[str] = None,
    image_author: Optional[str] = None,
    extra_dependencies: Optional[Tuple[Dependency, ...]] = None,
    dockerfile: Optional[Path] = None,
) -> None:
    """Build agent/service image."""
    extra_dependencies = extra_dependencies or ()
    if agent is None:
        service_dir = Path(service_dir or Path.cwd()).absolute()
        service = load_service_config(service_dir)
        agent = service.agent
        extra_dependencies = (*extra_dependencies, *service.dependencies.values())

    click.echo(f"Building image with agent: {agent}\n")
    _build_image(
        agent=agent,
        pull=pull,
        version=version,
        image_author=image_author,
        extra_dependencies=extra_dependencies,
        dockerfile=dockerfile,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the behaviours.py module of the skill."""
# pylint: skip-file

import platform
from abc import ABC
from calendar import timegm
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Tuple
from unittest import mock
from unittest.mock import MagicMock

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from packages.valory.skills.abstract_round_abci import PUBLIC_ID
from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    DegenerateRound,
    EventType,
    OffenseType,
    PendingOffense,
    RoundSequence,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    DegenerateBehaviour,
    TmManager,
)
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    PendingOffencesBehaviour,
    _MetaRoundBehaviour,
)
from packages.valory.skills.abstract_round_abci.models import TendermintRecoveryParams
from packages.valory.skills.abstract_round_abci.tests.conftest import profile_name


BEHAVIOUR_A_ID = "behaviour_a"
BEHAVIOUR_B_ID = "behaviour_b"
BEHAVIOUR_C_ID = "behaviour_c"
CONCRETE_BACKGROUND_BEHAVIOUR_ID = "background_behaviour"
ROUND_A_ID = "round_a"
ROUND_B_ID = "round_b"
CONCRETE_BACKGROUND_ROUND_ID = "background_round"


settings.load_profile(profile_name)


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


class RoundA(AbstractRound):
    """Round A."""

    round_id = ROUND_A_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class RoundB(AbstractRound):
    """Round B."""

    round_id = ROUND_B_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class ConcreteBackgroundRound(AbstractRound):
    """Concrete Background Round."""

    round_id = ROUND_B_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class BehaviourA(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_A_ID
    matching_round = RoundA

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize behaviour."""
        super().__init__(*args, **kwargs)
        self.count = 0

    def setup(self) -> None:
        """Setup behaviour."""
        self.count += 1

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class BehaviourB(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_B_ID
    matching_round = RoundB

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class BehaviourC(BaseBehaviour, ABC):
    """Dummy behaviour."""

    matching_round = MagicMock()


def test_auto_behaviour_id() -> None:
    """Test that the 'auto_behaviour_id()' method works as expected."""

    assert BehaviourB.auto_behaviour_id() == BEHAVIOUR_B_ID
    assert BehaviourB.behaviour_id == BEHAVIOUR_B_ID
    assert BehaviourC.auto_behaviour_id() == "behaviour_c"
    assert isinstance(BehaviourC.behaviour_id, property)


class ConcreteBackgroundBehaviour(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = CONCRETE_BACKGROUND_BEHAVIOUR_ID
    matching_round = ConcreteBackgroundRound

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class ConcreteAbciApp(AbciApp):
    """Concrete ABCI App."""

    initial_round_cls = RoundA
    transition_function = {RoundA: {MagicMock(): RoundB}}
    event_to_timeout: Dict = {}


class ConcreteRoundBehaviour(AbstractRoundBehaviour):
    """Concrete round behaviour."""

    abci_app_cls = ConcreteAbciApp
    behaviours = {BehaviourA, BehaviourB}  # type: ignore
    initial_behaviour_cls = BehaviourA
    background_behaviours_cls = {ConcreteBackgroundBehaviour}  # type: ignore


class TestAbstractRoundBehaviour:
    """Test 'AbstractRoundBehaviour' class."""

    def setup(self) -> None:
        """Set up the tests."""
        self.round_sequence_mock = MagicMock()
        context_mock = MagicMock(params=MagicMock())
        context_mock.state.round_sequence = self.round_sequence_mock
        context_mock.state.round_sequence.syncing_up = False
        self.round_sequence_mock.block_stall_deadline_expired = False
        self.behaviour = ConcreteRoundBehaviour(name="", skill_context=context_mock)

    @pytest.mark.parametrize("use_termination", (True, False))
    def test_setup(self, use_termination: bool) -> None:
        """Test 'setup' method."""
        assert self.behaviour.background_behaviours == set()
        self.behaviour.context.params.use_termination = use_termination
        self.behaviour.setup()
        assert self.behaviour.background_behaviours_cls == {ConcreteBackgroundBehaviour}
        assert (
            isinstance(
                self.behaviour.background_behaviours.pop(), ConcreteBackgroundBehaviour
            )
            if use_termination
            else self.behaviour.background_behaviours == set()
        )

    def test_teardown(self) -> None:
        """Test 'teardown' method."""
        self.behaviour.teardown()

    def test_current_behaviour_return_none(self) -> None:
        """Test 'current_behaviour' property return None."""
        assert self.behaviour.current_behaviour is None

    def test_act_current_behaviour_name_is_none(self) -> None:
        """Test 'act' with current behaviour None."""
        self.behaviour.tm_manager = self.behaviour.instantiate_behaviour_cls(TmManager)  # type: ignore
        self.behaviour.current_behaviour = None
        with mock.patch.object(self.behaviour, "_process_current_round"):
            self.behaviour.act()

    @pytest.mark.parametrize(
        "no_round, error",
        (
            (
                True,
                "Behaviour 'behaviour_without_round' specifies unknown 'unknown' as a matching round. "
                "Please make sure that the round is implemented and belongs to the FSM. "
                "If 'behaviour_without_round' is a background behaviour, please make sure that it is set correctly, "
                "by overriding the corresponding attribute of the chained skill's behaviour.",
            ),
            (False, "round round_1 is not a matching round of any behaviour"),
        ),
    )
    def test_check_matching_round_consistency_no_behaviour(
        self, no_round: bool, error: str
    ) -> None:
        """Test classmethod '_check_matching_round_consistency', when no behaviour or round is specified."""
        rounds = [
            MagicMock(**{"auto_round_id.return_value": f"round_{i}"}) for i in range(3)
        ]
        mock_behaviours = [
            MagicMock(matching_round=round, behaviour_id=f"behaviour_{i}")
            for i, round in enumerate(rounds[2:])
        ]
        if no_round:
            mock_behaviours.append(
                MagicMock(
                    matching_round="unknown", behaviour_id="behaviour_without_round"
                )
            )

        with mock.patch.object(
            _MetaRoundBehaviour, "_check_all_required_classattributes_are_set"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_behaviour_id_uniqueness"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_initial_behaviour_in_set_of_behaviours"
        ), pytest.raises(
            ABCIAppInternalError,
            match=error,
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock(
                    get_all_round_classes=lambda _, include_background_rounds: rounds,
                    final_states={
                        rounds[0],
                    },
                )
                behaviours = mock_behaviours  # type: ignore
                initial_behaviour_cls = MagicMock()

    def test_check_matching_round_consistency(self) -> None:
        """Test classmethod '_check_matching_round_consistency', negative case."""
        rounds = [
            MagicMock(**{"auto_round_id.return_value": f"round_{i}"}) for i in range(3)
        ]
        mock_behaviours = [
            MagicMock(matching_round=round, behaviour_id=f"behaviour_{i}")
            for i, round in enumerate(rounds)
        ]

        with mock.patch.object(
            _MetaRoundBehaviour, "_check_all_required_classattributes_are_set"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_behaviour_id_uniqueness"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_initial_behaviour_in_set_of_behaviours"
        ), pytest.raises(
            ABCIAppInternalError,
            match="internal error: round round_0 is a final round it shouldn't have any matching behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock(
                    get_all_round_classes=lambda _, include_background_rounds: rounds,
                    final_states={
                        rounds[0],
                    },
                )
                behaviours = mock_behaviours  # type: ignore
                initial_behaviour_cls = MagicMock()

    @pytest.mark.parametrize("behaviour_cls", (set(), {MagicMock()}))
    def test_check_matching_round_consistency_with_bg_rounds(
        self, behaviour_cls: set
    ) -> None:
        """Test classmethod '_check_matching_round_consistency' when a background behaviour class is set."""
        rounds = [
            MagicMock(**{"auto_round_id.return_value": f"round_{i}"}) for i in range(3)
        ]
        mock_behaviours = (
            [
                MagicMock(matching_round=round_, behaviour_id=f"behaviour_{i}")
                for i, round_ in enumerate(rounds[1:])
            ]
            if behaviour_cls
            else []
        )

        with mock.patch.object(
            _MetaRoundBehaviour, "_check_all_required_classattributes_are_set"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_behaviour_id_uniqueness"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_initial_behaviour_in_set_of_behaviours"
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock(
                    get_all_round_classes=lambda _, include_background_rounds: rounds
                    if include_background_rounds
                    else [],
                    final_states={
                        rounds[0],
                    }
                    if behaviour_cls
                    else {},
                )
                behaviours = mock_behaviours  # type: ignore
                initial_behaviour_cls = MagicMock()
                background_behaviours_cls = behaviour_cls

    def test_get_behaviour_id_to_behaviour_mapping_negative(self) -> None:
        """Test classmethod '_get_behaviour_id_to_behaviour_mapping', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1 = MagicMock(**{"auto_behaviour_id.return_value": behaviour_id})
        behaviour_2 = MagicMock(**{"auto_behaviour_id.return_value": behaviour_id})

        with pytest.raises(
            ValueError,
            match=f"cannot have two behaviours with the same id; got {behaviour_2} and {behaviour_1} both with id '{behaviour_id}'",
        ):
            with mock.patch.object(_MetaRoundBehaviour, "_check_consistency"):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = MagicMock
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = MagicMock()

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_two_behaviours_same_round(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.auto_round_id()
        behaviour_1 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": behaviour_id_1},
        )
        behaviour_2 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": behaviour_id_2},
        )

        with pytest.raises(
            ValueError,
            match=f"the behaviours '{behaviour_2.auto_behaviour_id()}' and '{behaviour_1.auto_behaviour_id()}' point to the same matching round '{round_id}'",
        ):
            with mock.patch.object(_MetaRoundBehaviour, "_check_consistency"):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = ConcreteAbciApp
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = behaviour_1

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_with_final_rounds(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' with final rounds."""

        class FinalRound(DegenerateRound, ABC):
            """A final round for testing."""

        behaviour_id_1 = "behaviour_id_1"
        behaviour_1 = MagicMock(behaviour_id=behaviour_id_1, matching_round=RoundA)

        class AbciAppTest(AbciApp):
            """Abci App for testing."""

            initial_round_cls = RoundA
            transition_function = {RoundA: {MagicMock(): FinalRound}, FinalRound: {}}
            event_to_timeout: Dict = {}
            final_states = {FinalRound}

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = AbciAppTest
            behaviours = {behaviour_1}
            initial_behaviour_cls = behaviour_1
            matching_round = FinalRound

        behaviour = MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())
        final_behaviour = behaviour._round_to_behaviour[FinalRound]
        assert issubclass(final_behaviour, DegenerateBehaviour)
        assert (
            final_behaviour.auto_behaviour_id()
            == f"degenerate_behaviour_{FinalRound.auto_round_id()}"
        )

    def test_check_behaviour_id_uniqueness_negative(self) -> None:
        """Test metaclass method '_check_consistency', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1_cls_name = "Behaviour1"
        behaviour_2_cls_name = "Behaviour2"
        behaviour_1 = MagicMock(
            __name__=behaviour_1_cls_name,
            **{"auto_behaviour_id.return_value": behaviour_id},
        )
        behaviour_2 = MagicMock(
            __name__=behaviour_2_cls_name,
            **{"auto_behaviour_id.return_value": behaviour_id},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=rf"behaviours \['{behaviour_1_cls_name}', '{behaviour_2_cls_name}'\] have the same behaviour id '{behaviour_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = MagicMock()

    def test_check_consistency_two_behaviours_same_round(self) -> None:
        """Test metaclass method '_check_consistency' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.auto_round_id()
        behaviour_1 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": "behaviour_id_1"},
        )
        behaviour_2 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": "behaviour_id_2"},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=rf"internal error: behaviours \['{behaviour_id_1}', '{behaviour_id_2}'\] have the same matching round '{round_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = behaviour_1

    def test_check_initial_behaviour_in_set_of_behaviours_negative_case(self) -> None:
        """Test classmethod '_check_initial_behaviour_in_set_of_behaviours' when initial behaviour is NOT in the set."""
        behaviour_1 = MagicMock(
            matching_round=MagicMock(),
            **{"auto_behaviour_id.return_value": "behaviour_id_1"},
        )
        behaviour_2 = MagicMock(
            matching_round=MagicMock(),
            **{"auto_behaviour_id.return_value": "behaviour_id_2"},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=f"initial behaviour {behaviour_2.auto_behaviour_id()} is not in the set of behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = {behaviour_1}
                initial_behaviour_cls = behaviour_2

    def test_act_no_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with no round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act(), current behaviour is initial behaviour and `clean_up()` has not been called
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # check that once the flag done is set, the `clean_up()` has been called
            # and `current_behaviour` is set to `None`.
            self.behaviour.current_behaviour.set_done()
            self.behaviour.act()
            assert self.behaviour.current_behaviour is None
            clean_up_mock.assert_called_once()

    def test_act_behaviour_setup(self) -> None:
        """Test the 'act' method of the FSM behaviour triggers setup() of the behaviour."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        assert self.behaviour.current_behaviour.count == 0

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act() first time, a call to setup has been made
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            assert self.behaviour.current_behaviour.count == 1

            # check that after act() second time, no further call to setup
            self.behaviour.act()
            assert self.behaviour.current_behaviour.count == 1

            # check that the `clean_up()` has not been called
            clean_up_mock.assert_not_called()

    def test_act_with_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that after act(), current behaviour is initial behaviour
        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # change the round
            self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
            self.round_sequence_mock.current_round_height = (
                self.round_sequence_mock.current_round_height + 1
            )

            # check that if the round is changed, the behaviour transition is performed and the clean-up is called
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourB)
            clean_up_mock.assert_called_once()

    def test_act_with_round_change_after_current_behaviour_is_none(self) -> None:
        """Test the 'act' method of the behaviour, with round change, after cur behaviour is none."""
        self.behaviour.tm_manager = self.behaviour.instantiate_behaviour_cls(TmManager)  # type: ignore
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # instantiate behaviour
        self.behaviour.current_behaviour = self.behaviour.instantiate_behaviour_cls(
            BehaviourA
        )

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act(), current behaviour is same behaviour
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # check that after the behaviour is done, current behaviour is None
            self.behaviour.current_behaviour.set_done()
            self.behaviour.act()
            assert self.behaviour.current_behaviour is None
            clean_up_mock.assert_called_once()

            # change the round
            self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
            self.round_sequence_mock.current_round_height = (
                self.round_sequence_mock.current_round_height + 1
            )

            # check that if the round is changed, the behaviour transition is taken
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourB)
            clean_up_mock.assert_called_once()

    @mock.patch.object(
        AbstractRoundBehaviour,
        "_process_current_round",
    )
    @mock.patch.object(
        TmManager,
        "tm_communication_unhealthy",
        new_callable=mock.PropertyMock,
        return_value=False,
    )
    @mock.patch.object(
        TmManager,
        "is_acting",
        new_callable=mock.PropertyMock,
        return_value=False,
    )
    @pytest.mark.parametrize("expected_termination_acting", (True, False))
    def test_termination_behaviour_acting(
        self,
        _: mock._patch,
        __: mock._patch,
        ___: mock._patch,
        expected_termination_acting: bool,
    ) -> None:
        """Test if the termination background behaviour is acting only when it should."""
        self.behaviour.context.params.use_termination = expected_termination_acting
        self.behaviour.setup()
        if expected_termination_acting:
            with mock.patch.object(
                ConcreteBackgroundBehaviour,
                "act_wrapper",
            ) as mock_background_act:
                self.behaviour.act()
                mock_background_act.assert_called()
        else:
            assert self.behaviour.background_behaviours == set()

    @mock.patch.object(
        AbstractRoundBehaviour,
        "_process_current_round",
    )
    @pytest.mark.parametrize(
        ("mock_tm_communication_unhealthy", "mock_is_acting", "expected_fix"),
        [
            (True, True, True),
            (False, True, True),
            (True, False, True),
            (False, False, False),
        ],
    )
    def test_try_fix_call(
        self,
        _: mock._patch,
        mock_tm_communication_unhealthy: bool,
        mock_is_acting: bool,
        expected_fix: bool,
    ) -> None:
        """Test that `try_fix` is called when necessary."""
        self.behaviour.tm_manager = self.behaviour.instantiate_behaviour_cls(TmManager)  # type: ignore
        with mock.patch.object(
            TmManager,
            "tm_communication_unhealthy",
            new_callable=mock.PropertyMock,
            return_value=mock_tm_communication_unhealthy,
        ), mock.patch.object(
            TmManager,
            "is_acting",
            new_callable=mock.PropertyMock,
            return_value=mock_is_acting,
        ), mock.patch.object(
            TmManager,
            "try_fix",
        ) as mock_try_fix:
            self.behaviour.act()
            if expected_fix:
                mock_try_fix.assert_called()
            else:
                mock_try_fix.assert_not_called()


def test_meta_round_behaviour_when_instance_not_subclass_of_abstract_round_behaviour() -> (
    None
):
    """Test instantiation of meta class when instance not a subclass of abstract round behaviour."""

    class MyRoundBehaviour(metaclass=_MetaRoundBehaviour):
        pass


def test_abstract_round_behaviour_instantiation_without_attributes_raises_error() -> (
    None
):
    """Test that definition of concrete subclass of AbstractRoundBehavior without attributes raises error."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            pass


def test_abstract_round_behaviour_matching_rounds_not_covered() -> None:
    """Test that definition of concrete subclass of AbstractRoundBehavior when matching round not covered."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = ConcreteAbciApp
            behaviours = {BehaviourA}
            initial_behaviour_cls = BehaviourA


@mock.patch.object(
    BaseBehaviour,
    "tm_communication_unhealthy",
    new_callable=mock.PropertyMock,
    return_value=False,
)
def test_self_loops_in_abci_app_reinstantiate_behaviour(_: mock._patch) -> None:
    """Test that a self-loop transition in the AbciApp will trigger a transition in the round behaviour."""
    event = MagicMock()

    class AbciAppTest(AbciApp):
        initial_round_cls = RoundA
        transition_function = {RoundA: {event: RoundA}}

    class RoundBehaviour(AbstractRoundBehaviour):
        abci_app_cls = AbciAppTest
        behaviours = {BehaviourA}
        initial_behaviour_cls = BehaviourA

    round_sequence = RoundSequence(MagicMock(), AbciAppTest)
    round_sequence.end_sync()
    round_sequence.setup(MagicMock(), MagicMock())
    context_mock = MagicMock()
    context_mock.state.round_sequence = round_sequence
    behaviour = RoundBehaviour(name="", skill_context=context_mock)
    behaviour.setup()

    behaviour_1 = behaviour.current_behaviour
    assert isinstance(behaviour_1, BehaviourA)

    round_sequence.abci_app.process_event(event)

    behaviour.act()
    behaviour_2 = behaviour.current_behaviour
    assert isinstance(behaviour_2, BehaviourA)
    assert id(behaviour_1) != id(behaviour_2)
    assert behaviour_1 != behaviour_2


class LongRunningBehaviour(BaseBehaviour):
    """A behaviour that runs forevever."""

    behaviour_id = "long_running_behaviour"
    matching_round = RoundA

    def async_act(self) -> Generator:
        """An act method that simply cycles forever."""
        while True:
            # cycle forever
            yield


def test_reset_should_be_performed_when_tm_unhealthy() -> None:
    """Test that hard reset is performed while a behaviour is running, and tendermint communication is unhealthy."""
    event = MagicMock()

    class AbciAppTest(AbciApp):
        initial_round_cls = RoundA
        transition_function = {RoundA: {event: RoundA}}

    class RoundBehaviour(AbstractRoundBehaviour):
        abci_app_cls = AbciAppTest
        behaviours = {LongRunningBehaviour}  # type: ignore
        initial_behaviour_cls = LongRunningBehaviour

    round_sequence = RoundSequence(MagicMock(), AbciAppTest)
    round_sequence.end_sync()
    round_sequence.setup(MagicMock(), MagicMock())
    context_mock = MagicMock()
    context_mock.state.round_sequence = round_sequence
    tm_recovery_params = TendermintRecoveryParams(
        reset_from_round=RoundA.auto_round_id()
    )
    context_mock.state.get_acn_result = MagicMock(return_value=tm_recovery_params)
    context_mock.params.ipfs_domain_name = None
    behaviour = RoundBehaviour(name="", skill_context=context_mock)
    behaviour.setup()

    current_behaviour = behaviour.current_behaviour
    assert isinstance(current_behaviour, LongRunningBehaviour)

    # upon entering the behaviour, the tendermint node communication is working well
    with mock.patch.object(
        RoundSequence,
        "block_stall_deadline_expired",
        new_callable=mock.PropertyMock,
        return_value=False,
    ):
        behaviour.act()

    def dummy_num_peers(
        timeout: Optional[float] = None,
    ) -> Generator[None, None, Optional[int]]:
        """A dummy method for num_active_peers."""
        # a None response is acceptable here, because tendermint is not healthy
        return None
        yield

    def dummy_reset_tendermint_with_wait(
        on_startup: bool = False,
        is_recovery: bool = False,
    ) -> Generator[None, None, bool]:
        """A dummy method for reset_tendermint_with_wait."""
        # we assume the reset goes through successfully
        return True
        yield

    # at this point LongRunningBehaviour is running
    # while the behaviour is running, the tendermint node
    # becomes unhealthy, we expect the node to be reset
    with mock.patch.object(
        RoundSequence,
        "block_stall_deadline_expired",
        new_callable=mock.PropertyMock,
        return_value=True,
    ), mock.patch.object(
        BaseBehaviour,
        "num_active_peers",
        side_effect=dummy_num_peers,
    ), mock.patch.object(
        BaseBehaviour,
        "reset_tendermint_with_wait",
        side_effect=dummy_reset_tendermint_with_wait,
    ) as mock_reset_tendermint:
        behaviour.tm_manager.synchronized_data.max_participants = 3  # type: ignore
        assert behaviour.tm_manager is not None
        behaviour.tm_manager.gentle_reset_attempted = True
        behaviour.act()
        mock_reset_tendermint.assert_called()


class TestPendingOffencesBehaviour:
    """Tests for `PendingOffencesBehaviour`."""

    behaviour: PendingOffencesBehaviour

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test class."""
        cls.behaviour = PendingOffencesBehaviour(
            name="test",
            skill_context=MagicMock(),
        )

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="`timegm` behaves differently on Windows. "
        "As a result, the generation of `last_transition_timestamp` is invalid.",
    )
    @given(
        offence=st.builds(
            PendingOffense,
            accused_agent_address=st.text(),
            round_count=st.integers(min_value=0),
            offense_type=st.sampled_from(OffenseType),
            last_transition_timestamp=st.floats(
                min_value=timegm(datetime(1971, 1, 1).utctimetuple()),
                max_value=timegm(datetime(8000, 1, 1).utctimetuple()) - 2000,
            ),
            time_to_live=st.floats(min_value=1, max_value=2000),
        ),
        wait_ticks=st.integers(min_value=0, max_value=1000),
        expired=st.booleans(),
    )
    def test_pending_offences_act(
        self,
        offence: PendingOffense,
        wait_ticks: int,
        expired: bool,
    ) -> None:
        """Test `PendingOffencesBehaviour`."""
        offence_expiration = offence.last_transition_timestamp + offence.time_to_live
        offence_expiration += 1 if expired else -1
        self.behaviour.round_sequence.last_round_transition_timestamp = datetime.fromtimestamp(  # type: ignore
            offence_expiration
        )

        gen = self.behaviour.async_act()

        with mock.patch.object(
            self.behaviour,
            "send_a2a_transaction",
        ) as mock_send_a2a_transaction, mock.patch.object(
            self.behaviour,
            "wait_until_round_end",
        ) as mock_wait_until_round_end, mock.patch.object(
            self.behaviour,
            "set_done",
        ) as mock_set_done:
            # while pending offences are empty, the behaviour simply waits
            for _ in range(wait_ticks):
                next(gen)

            self.behaviour.round_sequence.pending_offences = {offence}

            with pytest.raises(StopIteration):
                next(gen)

            check = "assert_not_called" if expired else "assert_called_once"

            for mocked in (
                mock_send_a2a_transaction,
                mock_wait_until_round_end,
                mock_set_done,
            ):
                getattr(mocked, check)()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/types/types.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2

from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)
from packages.valory.connections.abci.tendermint.crypto import (
    proof_pb2 as tendermint_dot_crypto_dot_proof__pb2,
)
from packages.valory.connections.abci.tendermint.types import (
    validator_pb2 as tendermint_dot_types_dot_validator__pb2,
)
from packages.valory.connections.abci.tendermint.version import (
    types_pb2 as tendermint_dot_version_dot_types__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1ctendermint/types/types.proto\x12\x10tendermint.types\x1a\x14gogoproto/gogo.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1dtendermint/crypto/proof.proto\x1a\x1etendermint/version/types.proto\x1a tendermint/types/validator.proto",\n\rPartSetHeader\x12\r\n\x05total\x18\x01 \x01(\r\x12\x0c\n\x04hash\x18\x02 \x01(\x0c"S\n\x04Part\x12\r\n\x05index\x18\x01 \x01(\r\x12\r\n\x05\x62ytes\x18\x02 \x01(\x0c\x12-\n\x05proof\x18\x03 \x01(\x0b\x32\x18.tendermint.crypto.ProofB\x04\xc8\xde\x1f\x00"W\n\x07\x42lockID\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12>\n\x0fpart_set_header\x18\x02 \x01(\x0b\x32\x1f.tendermint.types.PartSetHeaderB\x04\xc8\xde\x1f\x00"\xb3\x03\n\x06Header\x12\x34\n\x07version\x18\x01 \x01(\x0b\x32\x1d.tendermint.version.ConsensusB\x04\xc8\xde\x1f\x00\x12\x1d\n\x08\x63hain_id\x18\x02 \x01(\tB\x0b\xe2\xde\x1f\x07\x43hainID\x12\x0e\n\x06height\x18\x03 \x01(\x03\x12\x32\n\x04time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xc8\xde\x1f\x00\x90\xdf\x1f\x01\x12\x36\n\rlast_block_id\x18\x05 \x01(\x0b\x32\x19.tendermint.types.BlockIDB\x04\xc8\xde\x1f\x00\x12\x18\n\x10last_commit_hash\x18\x06 \x01(\x0c\x12\x11\n\tdata_hash\x18\x07 \x01(\x0c\x12\x17\n\x0fvalidators_hash\x18\x08 \x01(\x0c\x12\x1c\n\x14next_validators_hash\x18\t \x01(\x0c\x12\x16\n\x0e\x63onsensus_hash\x18\n \x01(\x0c\x12\x10\n\x08\x61pp_hash\x18\x0b \x01(\x0c\x12\x19\n\x11last_results_hash\x18\x0c \x01(\x0c\x12\x15\n\revidence_hash\x18\r \x01(\x0c\x12\x18\n\x10proposer_address\x18\x0e \x01(\x0c"\x13\n\x04\x44\x61ta\x12\x0b\n\x03txs\x18\x01 \x03(\x0c"\x92\x02\n\x04Vote\x12-\n\x04type\x18\x01 \x01(\x0e\x32\x1f.tendermint.types.SignedMsgType\x12\x0e\n\x06height\x18\x02 \x01(\x03\x12\r\n\x05round\x18\x03 \x01(\x05\x12<\n\x08\x62lock_id\x18\x04 \x01(\x0b\x32\x19.tendermint.types.BlockIDB\x0f\xc8\xde\x1f\x00\xe2\xde\x1f\x07\x42lockID\x12\x37\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xc8\xde\x1f\x00\x90\xdf\x1f\x01\x12\x19\n\x11validator_address\x18\x06 \x01(\x0c\x12\x17\n\x0fvalidator_index\x18\x07 \x01(\x05\x12\x11\n\tsignature\x18\x08 \x01(\x0c"\x9c\x01\n\x06\x43ommit\x12\x0e\n\x06height\x18\x01 \x01(\x03\x12\r\n\x05round\x18\x02 \x01(\x05\x12<\n\x08\x62lock_id\x18\x03 \x01(\x0b\x32\x19.tendermint.types.BlockIDB\x0f\xc8\xde\x1f\x00\xe2\xde\x1f\x07\x42lockID\x12\x35\n\nsignatures\x18\x04 \x03(\x0b\x32\x1b.tendermint.types.CommitSigB\x04\xc8\xde\x1f\x00"\xa8\x01\n\tCommitSig\x12\x34\n\rblock_id_flag\x18\x01 \x01(\x0e\x32\x1d.tendermint.types.BlockIDFlag\x12\x19\n\x11validator_address\x18\x02 \x01(\x0c\x12\x37\n\ttimestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xc8\xde\x1f\x00\x90\xdf\x1f\x01\x12\x11\n\tsignature\x18\x04 \x01(\x0c"\xf5\x01\n\x08Proposal\x12-\n\x04type\x18\x01 \x01(\x0e\x32\x1f.tendermint.types.SignedMsgType\x12\x0e\n\x06height\x18\x02 \x01(\x03\x12\r\n\x05round\x18\x03 \x01(\x05\x12\x11\n\tpol_round\x18\x04 \x01(\x05\x12<\n\x08\x62lock_id\x18\x05 \x01(\x0b\x32\x19.tendermint.types.BlockIDB\x0f\xc8\xde\x1f\x00\xe2\xde\x1f\x07\x42lockID\x12\x37\n\ttimestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xc8\xde\x1f\x00\x90\xdf\x1f\x01\x12\x11\n\tsignature\x18\x07 \x01(\x0c"b\n\x0cSignedHeader\x12(\n\x06header\x18\x01 \x01(\x0b\x32\x18.tendermint.types.Header\x12(\n\x06\x63ommit\x18\x02 \x01(\x0b\x32\x18.tendermint.types.Commit"z\n\nLightBlock\x12\x35\n\rsigned_header\x18\x01 \x01(\x0b\x32\x1e.tendermint.types.SignedHeader\x12\x35\n\rvalidator_set\x18\x02 \x01(\x0b\x32\x1e.tendermint.types.ValidatorSet"\x9e\x01\n\tBlockMeta\x12<\n\x08\x62lock_id\x18\x01 \x01(\x0b\x32\x19.tendermint.types.BlockIDB\x0f\xc8\xde\x1f\x00\xe2\xde\x1f\x07\x42lockID\x12\x12\n\nblock_size\x18\x02 \x01(\x03\x12.\n\x06header\x18\x03 \x01(\x0b\x32\x18.tendermint.types.HeaderB\x04\xc8\xde\x1f\x00\x12\x0f\n\x07num_txs\x18\x04 \x01(\x03"S\n\x07TxProof\x12\x11\n\troot_hash\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\'\n\x05proof\x18\x03 \x01(\x0b\x32\x18.tendermint.crypto.Proof*\xd7\x01\n\x0b\x42lockIDFlag\x12\x31\n\x15\x42LOCK_ID_FLAG_UNKNOWN\x10\x00\x1a\x16\x8a\x9d \x12\x42lockIDFlagUnknown\x12/\n\x14\x42LOCK_ID_FLAG_ABSENT\x10\x01\x1a\x15\x8a\x9d \x11\x42lockIDFlagAbsent\x12/\n\x14\x42LOCK_ID_FLAG_COMMIT\x10\x02\x1a\x15\x8a\x9d \x11\x42lockIDFlagCommit\x12)\n\x11\x42LOCK_ID_FLAG_NIL\x10\x03\x1a\x12\x8a\x9d \x0e\x42lockIDFlagNil\x1a\x08\x88\xa3\x1e\x00\xa8\xa4\x1e\x01*\xd7\x01\n\rSignedMsgType\x12,\n\x17SIGNED_MSG_TYPE_UNKNOWN\x10\x00\x1a\x0f\x8a\x9d \x0bUnknownType\x12,\n\x17SIGNED_MSG_TYPE_PREVOTE\x10\x01\x1a\x0f\x8a\x9d \x0bPrevoteType\x12\x30\n\x19SIGNED_MSG_TYPE_PRECOMMIT\x10\x02\x1a\x11\x8a\x9d \rPrecommitType\x12.\n\x18SIGNED_MSG_TYPE_PROPOSAL\x10 \x1a\x10\x8a\x9d \x0cProposalType\x1a\x08\x88\xa3\x1e\x00\xa8\xa4\x1e\x01\x42\x39Z7github.com/tendermint/tendermint/proto/tendermint/typesb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "tendermint.types.types_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z7github.com/tendermint/tendermint/proto/tendermint/types"
    )
    _BLOCKIDFLAG._options = None
    _BLOCKIDFLAG._serialized_options = b"\210\243\036\000\250\244\036\001"
    _BLOCKIDFLAG.values_by_name["BLOCK_ID_FLAG_UNKNOWN"]._options = None
    _BLOCKIDFLAG.values_by_name[
        "BLOCK_ID_FLAG_UNKNOWN"
    ]._serialized_options = b"\212\235 \022BlockIDFlagUnknown"
    _BLOCKIDFLAG.values_by_name["BLOCK_ID_FLAG_ABSENT"]._options = None
    _BLOCKIDFLAG.values_by_name[
        "BLOCK_ID_FLAG_ABSENT"
    ]._serialized_options = b"\212\235 \021BlockIDFlagAbsent"
    _BLOCKIDFLAG.values_by_name["BLOCK_ID_FLAG_COMMIT"]._options = None
    _BLOCKIDFLAG.values_by_name[
        "BLOCK_ID_FLAG_COMMIT"
    ]._serialized_options = b"\212\235 \021BlockIDFlagCommit"
    _BLOCKIDFLAG.values_by_name["BLOCK_ID_FLAG_NIL"]._options = None
    _BLOCKIDFLAG.values_by_name[
        "BLOCK_ID_FLAG_NIL"
    ]._serialized_options = b"\212\235 \016BlockIDFlagNil"
    _SIGNEDMSGTYPE._options = None
    _SIGNEDMSGTYPE._serialized_options = b"\210\243\036\000\250\244\036\001"
    _SIGNEDMSGTYPE.values_by_name["SIGNED_MSG_TYPE_UNKNOWN"]._options = None
    _SIGNEDMSGTYPE.values_by_name[
        "SIGNED_MSG_TYPE_UNKNOWN"
    ]._serialized_options = b"\212\235 \013UnknownType"
    _SIGNEDMSGTYPE.values_by_name["SIGNED_MSG_TYPE_PREVOTE"]._options = None
    _SIGNEDMSGTYPE.values_by_name[
        "SIGNED_MSG_TYPE_PREVOTE"
    ]._serialized_options = b"\212\235 \013PrevoteType"
    _SIGNEDMSGTYPE.values_by_name["SIGNED_MSG_TYPE_PRECOMMIT"]._options = None
    _SIGNEDMSGTYPE.values_by_name[
        "SIGNED_MSG_TYPE_PRECOMMIT"
    ]._serialized_options = b"\212\235 \rPrecommitType"
    _SIGNEDMSGTYPE.values_by_name["SIGNED_MSG_TYPE_PROPOSAL"]._options = None
    _SIGNEDMSGTYPE.values_by_name[
        "SIGNED_MSG_TYPE_PROPOSAL"
    ]._serialized_options = b"\212\235 \014ProposalType"
    _PART.fields_by_name["proof"]._options = None
    _PART.fields_by_name["proof"]._serialized_options = b"\310\336\037\000"
    _BLOCKID.fields_by_name["part_set_header"]._options = None
    _BLOCKID.fields_by_name["part_set_header"]._serialized_options = b"\310\336\037\000"
    _HEADER.fields_by_name["version"]._options = None
    _HEADER.fields_by_name["version"]._serialized_options = b"\310\336\037\000"
    _HEADER.fields_by_name["chain_id"]._options = None
    _HEADER.fields_by_name["chain_id"]._serialized_options = b"\342\336\037\007ChainID"
    _HEADER.fields_by_name["time"]._options = None
    _HEADER.fields_by_name[
        "time"
    ]._serialized_options = b"\310\336\037\000\220\337\037\001"
    _HEADER.fields_by_name["last_block_id"]._options = None
    _HEADER.fields_by_name["last_block_id"]._serialized_options = b"\310\336\037\000"
    _VOTE.fields_by_name["block_id"]._options = None
    _VOTE.fields_by_name[
        "block_id"
    ]._serialized_options = b"\310\336\037\000\342\336\037\007BlockID"
    _VOTE.fields_by_name["timestamp"]._options = None
    _VOTE.fields_by_name[
        "timestamp"
    ]._serialized_options = b"\310\336\037\000\220\337\037\001"
    _COMMIT.fields_by_name["block_id"]._options = None
    _COMMIT.fields_by_name[
        "block_id"
    ]._serialized_options = b"\310\336\037\000\342\336\037\007BlockID"
    _COMMIT.fields_by_name["signatures"]._options = None
    _COMMIT.fields_by_name["signatures"]._serialized_options = b"\310\336\037\000"
    _COMMITSIG.fields_by_name["timestamp"]._options = None
    _COMMITSIG.fields_by_name[
        "timestamp"
    ]._serialized_options = b"\310\336\037\000\220\337\037\001"
    _PROPOSAL.fields_by_name["block_id"]._options = None
    _PROPOSAL.fields_by_name[
        "block_id"
    ]._serialized_options = b"\310\336\037\000\342\336\037\007BlockID"
    _PROPOSAL.fields_by_name["timestamp"]._options = None
    _PROPOSAL.fields_by_name[
        "timestamp"
    ]._serialized_options = b"\310\336\037\000\220\337\037\001"
    _BLOCKMETA.fields_by_name["block_id"]._options = None
    _BLOCKMETA.fields_by_name[
        "block_id"
    ]._serialized_options = b"\310\336\037\000\342\336\037\007BlockID"
    _BLOCKMETA.fields_by_name["header"]._options = None
    _BLOCKMETA.fields_by_name["header"]._serialized_options = b"\310\336\037\000"
    _globals["_BLOCKIDFLAG"]._serialized_start = 2207
    _globals["_BLOCKIDFLAG"]._serialized_end = 2422
    _globals["_SIGNEDMSGTYPE"]._serialized_start = 2425
    _globals["_SIGNEDMSGTYPE"]._serialized_end = 2640
    _globals["_PARTSETHEADER"]._serialized_start = 202
    _globals["_PARTSETHEADER"]._serialized_end = 246
    _globals["_PART"]._serialized_start = 248
    _globals["_PART"]._serialized_end = 331
    _globals["_BLOCKID"]._serialized_start = 333
    _globals["_BLOCKID"]._serialized_end = 420
    _globals["_HEADER"]._serialized_start = 423
    _globals["_HEADER"]._serialized_end = 858
    _globals["_DATA"]._serialized_start = 860
    _globals["_DATA"]._serialized_end = 879
    _globals["_VOTE"]._serialized_start = 882
    _globals["_VOTE"]._serialized_end = 1156
    _globals["_COMMIT"]._serialized_start = 1159
    _globals["_COMMIT"]._serialized_end = 1315
    _globals["_COMMITSIG"]._serialized_start = 1318
    _globals["_COMMITSIG"]._serialized_end = 1486
    _globals["_PROPOSAL"]._serialized_start = 1489
    _globals["_PROPOSAL"]._serialized_end = 1734
    _globals["_SIGNEDHEADER"]._serialized_start = 1736
    _globals["_SIGNEDHEADER"]._serialized_end = 1834
    _globals["_LIGHTBLOCK"]._serialized_start = 1836
    _globals["_LIGHTBLOCK"]._serialized_end = 1958
    _globals["_BLOCKMETA"]._serialized_start = 1961
    _globals["_BLOCKMETA"]._serialized_end = 2119
    _globals["_TXPROOF"]._serialized_start = 2121
    _globals["_TXPROOF"]._serialized_end = 2204
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `run` command."""


import json
import os
import shutil
from unittest import mock

from aea.configurations.constants import DEFAULT_AEA_CONFIG_FILE, PACKAGES

from autonomy.constants import DEFAULT_BUILD_FOLDER, DOCKER_COMPOSE_YAML, VALORY
from autonomy.deploy.base import ServiceBuilder
from autonomy.deploy.constants import (
    AGENT_VARS_CONFIG_FILE,
    TENDERMINT_VARS_CONFIG_FILE,
)

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest
from tests.test_autonomy.test_cli.test_deploy.test_build.test_deployment import (
    OS_ENV_PATCH,
)


class TestRun(BaseCliTest):
    """Test `run` command."""

    cli_options = ("deploy", "run")

    def setup(self) -> None:
        """Setup test method."""

        super().setup()
        os.chdir(self.t)

    def test_run(
        self,
    ) -> None:
        """Run test."""
        (self.t / DOCKER_COMPOSE_YAML).touch()
        with mock.patch(
            "autonomy.cli.helpers.deployment.docker_compose.project_from_options"
        ), mock.patch("autonomy.cli.helpers.deployment.docker_compose.TopLevelCommand"):
            result = self.run_cli(("--detach",))
            assert result.exit_code == 0, result.output
            assert "Running build @" in result.output

    def test_run_local(self) -> None:
        """Test that `deploy run` works on localhost."""
        super().setup()

        # setup the service keys and packages
        self.keys_file = self.t / "keys.json"
        shutil.copytree(ROOT_DIR / PACKAGES, self.t / PACKAGES)
        shutil.copy(
            ROOT_DIR / "deployments" / "keys" / "hardhat_keys.json", self.keys_file
        )
        shutil.copytree(
            self.t / PACKAGES / "valory" / "services" / "register_reset",
            self.t / "register_reset",
        )
        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset",
                self.keys_file,
            )
        os.chdir(self.t / "register_reset")

        # setup aea-config.yaml
        shutil.copy(
            ROOT_DIR
            / PACKAGES
            / VALORY
            / "agents"
            / "register_reset"
            / DEFAULT_AEA_CONFIG_FILE,
            self.t / "register_reset",
        )

        # setup agent.json and tendermint.json
        os.mkdir(build_path := self.t / "register_reset" / DEFAULT_BUILD_FOLDER)
        with open(build_path / TENDERMINT_VARS_CONFIG_FILE, "w") as fp:
            json.dump({}, fp)
        with open(build_path / AGENT_VARS_CONFIG_FILE, "w") as fp:
            json.dump({}, fp)
        with mock.patch("autonomy.cli.helpers.deployment.subprocess.run"), mock.patch(
            "autonomy.cli.helpers.deployment.subprocess.Popen"
        ), mock.patch("autonomy.cli.helpers.deployment.check_tendermint_version"):
            result = self.run_cli(("--localhost", "--build-dir", build_path.as_posix()))
            assert result.exit_code == 0, result.output
            assert "Running build @" in result.output

    def test_missing_config_file(
        self,
    ) -> None:
        """Run test."""
        result = self.run_cli()
        assert result.exit_code == 1, result.output
        assert "Deployment configuration does not exist" in result.output


class TestStop(BaseCliTest):
    """Test `stop` command."""

    cli_options = ("deploy", "stop")

    def setup(self) -> None:
        """Setup test method."""
        super().setup()
        os.chdir(self.t)

    def test_run(
        self,
    ) -> None:
        """Run test."""
        (self.t / DOCKER_COMPOSE_YAML).touch()
        with mock.patch(
            "autonomy.cli.helpers.deployment.docker_compose.project_from_options"
        ), mock.patch("autonomy.cli.helpers.deployment.docker_compose.TopLevelCommand"):
            result = self.run_cli()
            assert result.exit_code == 0, result.output
            assert "Don't cancel while stopping services..." in result.output

    def test_missing_config_file(
        self,
    ) -> None:
        """Run test."""
        result = self.run_cli()
        assert result.exit_code == 1, result.output
        assert "Deployment configuration does not exist" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""On-chain tools configurations."""

import os
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, cast

from autonomy.chain.constants import CHAIN_NAME_TO_CHAIN_ID, CHAIN_PROFILES


DEFAULT_LOCAL_RPC = "http://127.0.0.1:8545"
DEFAULT_LOCAL_CHAIN_ID = 31337
CUSTOM_CHAIN_RPC = "CUSTOM_CHAIN_RPC"
ETHEREUM_CHAIN_RPC = "ETHEREUM_CHAIN_RPC"


def _get_chain_id_for_custom_chain() -> Optional[int]:
    """Get chain if for custom chain from environment"""
    chain_id = os.environ.get("CUSTOM_CHAIN_ID")
    if chain_id is None:
        return None
    return int(chain_id)


class ChainType(Enum):
    """Chain types."""

    LOCAL = "local"
    CUSTOM = "custom_chain"
    ETHEREUM = "ethereum"
    POLYGON = "polygon"
    GNOSIS = "gnosis"
    ARBITRUM_ONE = "arbitrum_one"
    OPTIMISTIC = "optimistic"
    BASE = "base"
    CELO = "celo"
    MODE = "mode"

    @property
    def id(self) -> Optional[int]:
        """Chain ID"""
        if self == ChainType.CUSTOM:
            return _get_chain_id_for_custom_chain()
        return CHAIN_NAME_TO_CHAIN_ID[self.value]

    @property
    def rpc(self) -> Optional[str]:
        """RPC String"""
        if self == ChainType.LOCAL:
            return DEFAULT_LOCAL_RPC
        return os.environ.get(self.rpc_env_name)

    @property
    def rpc_env_name(self) -> str:
        """RPC Environment variable name"""
        if self == ChainType.CUSTOM:
            return f"{self.value}_rpc".upper()
        return f"{self.value}_chain_rpc".upper()


Chain = ChainType


@dataclass
class ContractConfig:
    """Contract config class."""

    name: str
    contracts: Dict[ChainType, str]


@dataclass
class ChainConfig:
    """Chain config"""

    chain_type: ChainType
    rpc: Optional[str]
    chain_id: Optional[int]


class ChainConfigs:  # pylint: disable=too-few-public-methods
    """Name space for chain configs."""

    local = ChainConfig(
        chain_type=ChainType.LOCAL,
        rpc=DEFAULT_LOCAL_RPC,
        chain_id=DEFAULT_LOCAL_CHAIN_ID,
    )

    @classmethod
    def get(cls, chain_type: ChainType) -> ChainConfig:
        """Return chain config for given chain type."""
        return ChainConfig(
            chain_type=chain_type,
            rpc=chain_type.rpc,
            chain_id=chain_type.id,
        )

    @classmethod
    def get_rpc_env_var(cls, chain_type: ChainType) -> Optional[str]:
        """Return chain config for given chain type."""
        return chain_type.rpc_env_name


class ContractConfigs:  # pylint: disable=too-few-public-methods
    """A namespace for contract configs."""

    component_registry = ContractConfig(
        name="component_registry",
        contracts={
            ChainType(chain_name): cast(str, container.get("component_registry"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    agent_registry = ContractConfig(
        name="agent_registry",
        contracts={
            ChainType(chain_name): cast(str, container.get("agent_registry"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    service_registry = ContractConfig(
        name="service_registry",
        contracts={
            ChainType(chain_name): cast(str, container.get("service_registry"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    service_manager = ContractConfig(
        name="service_manager",
        contracts={
            ChainType(chain_name): cast(
                str,
                container.get(
                    "service_manager", container.get("service_manager_token")
                ),
            )
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    registries_manager = ContractConfig(
        name="registries_manager",
        contracts={
            ChainType(chain_name): cast(str, container.get("registries_manager"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    gnosis_safe_proxy_factory = ContractConfig(
        name="gnosis_safe_proxy_factory",
        contracts={
            ChainType(chain_name): cast(str, container.get("gnosis_safe_proxy_factory"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    gnosis_safe_same_address_multisig = ContractConfig(
        name="gnosis_safe_same_address_multisig",
        contracts={
            ChainType(chain_name): cast(
                str, container.get("gnosis_safe_same_address_multisig")
            )
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    service_registry_token_utility = ContractConfig(
        name="service_registry_token_utility",
        contracts={
            ChainType(chain_name): cast(
                str, container.get("service_registry_token_utility")
            )
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    multisend = ContractConfig(
        name="multisend",
        contracts={
            ChainType(chain_name): cast(str, container.get("multisend"))
            for chain_name, container in CHAIN_PROFILES.items()
        },
    )

    erc20 = ContractConfig(
        name="erc20",
        contracts={},
    )

    @classmethod
    def get(cls, name: str) -> ContractConfig:
        """Return chain config for given chain type."""
        return cast(ContractConfig, getattr(cls, name))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Constants for generating deployments environment."""

from pathlib import Path
from string import Template


DEPLOYMENT_REPORT: Template = Template(
    """
Generated Deployment!\n\n
Type:                 $type
Agents:               $agents
Build Length          $size\n\n
"""
)

DEPLOYMENT_KEY_DIRECTORY = "agent_keys"
DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA = "agent_{agent_n}"
KUBERNETES_AGENT_KEY_NAME = DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA + "_private_key.yaml"
TENDERMINT_BIN_UNIX = "tendermint"
TENDERMINT_BIN_WINDOWS = "tendermint.exe"
TENDERMINT_VARS_CONFIG_FILE = "tendermint.json"
AGENT_VARS_CONFIG_FILE = "agent.json"
TENDERMINT_FLASK_APP_PATH = (
    Path("autonomy") / "deploy" / "generators" / "localhost" / "tendermint" / "app.py"
)
DEATTACH_WINDOWS_FLAG = 0x00000008

TM_ENV_TMHOME = "TMHOME"
TM_ENV_TMSTATE = "TMSTATE"
TM_ENV_PROXY_APP = "PROXY_APP"
TM_ENV_P2P_LADDR = "P2P_LADDR"
TM_ENV_RPC_LADDR = "RPC_LADDR"
TM_ENV_PROXY_APP = "PROXY_APP"
TM_ENV_CREATE_EMPTY_BLOCKS = "CREATE_EMPTY_BLOCKS"
TM_ENV_USE_GRPC = "USE_GRPC"

DEFAULT_ENCODING = "utf-8"

PERSISTENT_DATA_DIR = "persistent_data"
LOG_DIR = "logs"
TM_STATE_DIR = "tm_state"
BENCHMARKS_DIR = "benchmarks"
VENVS_DIR = "venvs"
AGENT_KEYS_DIR = "agent_keys"
DOCKERFILES = "Dockerfiles"

INFO = "INFO"
DEBUG = "DEBUG"
WARNING = "WARNING"
ERROR = "ERROR"
CRITICAL = "CRITICAL"

LOGGING_LEVELS = (
    INFO,
    DEBUG,
    WARNING,
    ERROR,
    CRITICAL,
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test log parser."""

import tempfile
from pathlib import Path

from autonomy.analyse.logs.collection import LogCollection


LOGS = """[2023-09-26 06:27:56,015] [INFO] [agent] Entered in the 'check_transaction_history_behaviour' behaviour
[2023-09-26 06:27:56,078] [ERROR] [agent] get_safe_nonce unsuccessful! Received: Message(sender=valory/ledger:0.19.0,to=valory/trader_abci:0.1.0,code=500,data=b'',dialogue_reference=('974278ff7b4e00019f7542a193987a8359b1c785d11a98a84bfaff2ea77b1e8f', '6b8e33fe183de2ac45e8d8f9625db5562b6230677fdc6995efc4ce05c798f83a'),message=Traceback (most recent call last):

  File "/usr/lib/python3.10/http/client.py", line 287, in _read_status
    raise RemoteDisconnected("Remote end closed connection without"

http.client.RemoteDisconnected: Remote end closed connection without response


During handling of the above exception, another exception occurred:


Traceback (most recent call last):

  File "/usr/lib/python3.10/http/client.py", line 287, in _read_status
    raise RemoteDisconnected("Remote end closed connection without"

urllib3.exceptions.ProtocolError: ('Connection aborted.', RemoteDisconnected('Remote end closed co,message_id=-1,performative=error,target=1)
[2023-09-26 06:27:56,103] [INFO] [agent] arrived block with timestamp: 2023-09-26 06:27:55.040545
"""

LOGS_CLEAN = """Entered in the 'check_transaction_history_behaviour' behaviour
get_safe_nonce unsuccessful! Received: Message(sender=valory/ledger:0.19.0,to=valory/trader_abci:0.1.0,code=500,data=b'',dialogue_reference=('974278ff7b4e00019f7542a193987a8359b1c785d11a98a84bfaff2ea77b1e8f', '6b8e33fe183de2ac45e8d8f9625db5562b6230677fdc6995efc4ce05c798f83a'),message=Traceback (most recent call last):
File "/usr/lib/python3.10/http/client.py", line 287, in _read_status
raise RemoteDisconnected("Remote end closed connection without"
http.client.RemoteDisconnected: Remote end closed connection without response
During handling of the above exception, another exception occurred:
Traceback (most recent call last):
File "/usr/lib/python3.10/http/client.py", line 287, in _read_status
raise RemoteDisconnected("Remote end closed connection without"
urllib3.exceptions.ProtocolError: ('Connection aborted.', RemoteDisconnected('Remote end closed co,message_id=-1,performative=error,target=1)
arrived block with timestamp: 2023-09-26 06:27:55.040545
"""


def test_multiline_parse() -> None:
    """Test multiline logs parsing."""

    with tempfile.TemporaryDirectory() as temp_dir:
        file = Path(temp_dir, "log.txt")
        file.write_text(LOGS)
        parsed = "\n".join(list(map(lambda x: x[2], LogCollection.parse(file=file))))

        for line in LOGS_CLEAN.split("\n"):
            assert line in parsed



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Verify the on-chain addresses against the defined constants."""

import typing as t

import pytest
import requests
from aea.protocols.generator.common import _camel_case_to_snake_case

from autonomy.chain.config import ChainType
from autonomy.chain.constants import CHAIN_PROFILES


ADDRESS_FILE_URL = "https://raw.githubusercontent.com/valory-xyz/autonolas-registries/refs/tags/v1.2.4/docs/configuration.json"


class TestAddresses:
    """Test addresses."""

    contracts: t.Dict[str, t.List[t.Dict[str, str]]]

    @classmethod
    def setup_class(cls) -> None:
        """Setup test class."""
        chain_configs = requests.get(url=ADDRESS_FILE_URL).json()
        cls.contracts = {
            _camel_case_to_snake_case(config["name"]): config["contracts"]
            for config in chain_configs
        }

    @pytest.mark.parametrize(argnames="chain", argvalues=list(ChainType))
    def test_addresses_match(self, chain: ChainType) -> None:
        """Test addresses match with the remote file."""
        if chain in (ChainType.LOCAL, ChainType.CUSTOM):
            return

        if chain == ChainType.ETHEREUM:
            contracts = self.contracts["mainnet"]
        else:
            contracts = self.contracts[chain.value]

        for contract in contracts:
            name = _camel_case_to_snake_case(contract["name"]).replace("_l2", "")
            address = contract["address"]
            if name == "gnosis_safe_multisig":
                constant_address = CHAIN_PROFILES[chain.value][
                    "gnosis_safe_proxy_factory"
                ]
            else:
                constant_address = CHAIN_PROFILES[chain.value][name]
            assert (
                address == constant_address
            ), f"Constant value and remote value does not match for `{name}`"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test deployment command."""

import json
import os
import shutil
from itertools import cycle
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from unittest import mock

import yaml
from aea.cli.registry.settings import REGISTRY_LOCAL
from aea.cli.utils.config import get_default_author_from_cli_config
from aea.cli.utils.constants import CLI_CONFIG_PATH, DEFAULT_CLI_CONFIG
from aea.configurations.constants import (
    CONNECTION,
    CONTRACT,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_ENV_DOTFILE,
    PACKAGES,
    PROTOCOL,
    SKILL,
)
from aea_test_autonomy.configurations import (
    ETHEREUM_ENCRYPTED_KEYS,
    ETHEREUM_ENCRYPTION_PASSWORD,
)

from autonomy.constants import (
    DEFAULT_BUILD_FOLDER,
    DEFAULT_DOCKER_IMAGE_AUTHOR,
    DOCKER_COMPOSE_YAML,
    VALORY,
)
from autonomy.deploy.base import (
    DEFAULT_AGENT_CPU_LIMIT,
    DEFAULT_AGENT_CPU_REQUEST,
    DEFAULT_AGENT_MEMORY_LIMIT,
    DEFAULT_AGENT_MEMORY_REQUEST,
    ServiceBuilder,
    build_hash_id,
)
from autonomy.deploy.constants import (
    AGENT_VARS_CONFIG_FILE,
    DEBUG,
    DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA,
    DEPLOYMENT_KEY_DIRECTORY,
    INFO,
    KUBERNETES_AGENT_KEY_NAME,
    PERSISTENT_DATA_DIR,
    TENDERMINT_VARS_CONFIG_FILE,
    TM_ENV_CREATE_EMPTY_BLOCKS,
    TM_ENV_P2P_LADDR,
    TM_ENV_PROXY_APP,
    TM_ENV_RPC_LADDR,
    TM_ENV_TMHOME,
    TM_ENV_TMSTATE,
    TM_ENV_USE_GRPC,
    TM_STATE_DIR,
)
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.localhost.utils import _run_aea_cmd
from autonomy.replay.agent import ETHEREUM_PRIVATE_KEY_FILE

from tests.conftest import ROOT_DIR, skip_docker_tests
from tests.test_autonomy.base import get_dummy_service_config
from tests.test_autonomy.test_cli.base import BaseCliTest


OS_ENV_PATCH = mock.patch.dict(
    os.environ, values={**os.environ, "ALL_PARTICIPANTS": "[]"}, clear=True
)


@skip_docker_tests
class BaseDeployBuildTest(BaseCliTest):
    """Test `autonomy deply build deployment` command."""

    cli_options: Tuple[str, ...] = ("deploy", "build")
    keys_file: Path
    spec: ServiceBuilder

    def setup(self) -> None:
        """Setup test method."""
        super().setup()

        self.keys_file = self.t / "keys.json"
        shutil.copytree(ROOT_DIR / PACKAGES, self.t / PACKAGES)
        shutil.copy(
            ROOT_DIR / "deployments" / "keys" / "hardhat_keys.json", self.keys_file
        )

        shutil.copytree(
            self.t / PACKAGES / "valory" / "services" / "register_reset",
            self.t / "register_reset",
        )
        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset", self.keys_file, service_hash_id="test"
            )
        os.chdir(self.t / "register_reset")

    @staticmethod
    def load_kubernetes_config(
        path: Path,
    ) -> List[Dict]:
        """Load kubernetes config."""

        with open(path / "build.yaml", "r") as fp:
            return list(yaml.safe_load_all(fp))

    @classmethod
    def check_kubernetes_build(cls, build_dir: Path) -> None:
        """Check kubernetes build dir."""

        build_tree = list(map(lambda x: x.name, build_dir.iterdir()))
        assert any(
            child in build_tree for child in ["persistent_storage", "build.yaml"]
        )

    def load_and_check_docker_compose_file(
        self,
        path: Path,
        service_hash_id: Optional[str] = None,
    ) -> Dict:
        """Load docker compose config."""
        with open(path, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        if service_hash_id is not None:
            with OS_ENV_PATCH:
                self.spec = ServiceBuilder.from_dir(
                    self.t / "register_reset",
                    self.keys_file,
                    service_hash_id=service_hash_id,
                )

        assert any(
            [key in docker_compose for key in ["version", "services", "networks"]]
        )

        assert any(
            [
                service in docker_compose["services"]
                for service in [
                    *map(lambda i: self.spec.get_abci_container_name(i), range(4)),
                    *map(lambda i: self.spec.get_tm_container_name(i), range(4)),
                ]
            ]
        )

        return docker_compose

    @staticmethod
    def check_localhost_build(build_dir: Path) -> None:
        """Check localhost build directory."""
        build_tree = list(map(lambda x: x.name, build_dir.iterdir()))
        assert all(
            [
                child in build_tree
                for child in {
                    ".certs",
                    DEFAULT_AEA_CONFIG_FILE,
                    DEPLOYMENT_KEY_DIRECTORY,
                    ETHEREUM_PRIVATE_KEY_FILE,
                    PERSISTENT_DATA_DIR,
                    AGENT_VARS_CONFIG_FILE,
                    "data",
                    "node",
                    TENDERMINT_VARS_CONFIG_FILE,
                }
            ]
        )

    def load_and_check_localhost_build(self, path: Path) -> None:
        """Load localhost build config."""
        with open(path / TENDERMINT_VARS_CONFIG_FILE, "r", encoding="utf-8") as fp:
            assert json.load(fp) == {
                TM_ENV_TMHOME: (
                    self.t / "register_reset" / DEFAULT_BUILD_FOLDER / "node"
                ).as_posix(),
                TM_ENV_TMSTATE: (
                    self.t / "register_reset" / DEFAULT_BUILD_FOLDER / TM_STATE_DIR
                ).as_posix(),
                TM_ENV_P2P_LADDR: "tcp://localhost:26656",
                TM_ENV_RPC_LADDR: "tcp://localhost:26657",
                TM_ENV_PROXY_APP: "tcp://localhost:26658",
                TM_ENV_CREATE_EMPTY_BLOCKS: "true",
                TM_ENV_USE_GRPC: "false",
            }
        with open(path / AGENT_VARS_CONFIG_FILE, "r", encoding="utf-8") as fp:
            assert json.load(fp) == {
                "ID": "0",
                "AEA_AGENT": "valory/register_reset:0.1.0:bafybeia4pxlphcvco3ttlv3tytklriwvuwxxxr5m2tdry32yc5vogxtm7u",
                "LOG_LEVEL": INFO,
                "AEA_PASSWORD": "",
                "CONNECTION_LEDGER_CONFIG_LEDGER_APIS_ETHEREUM_ADDRESS": "http://host.docker.internal:8545",
                "CONNECTION_LEDGER_CONFIG_LEDGER_APIS_ETHEREUM_CHAIN_ID": "31337",
                "CONNECTION_LEDGER_CONFIG_LEDGER_APIS_ETHEREUM_POA_CHAIN": "False",
                "CONNECTION_LEDGER_CONFIG_LEDGER_APIS_ETHEREUM_DEFAULT_GAS_PRICE_STRATEGY": "eip1559",
                "CONNECTION_ABCI_CONFIG_HOST": "localhost",
                "CONNECTION_ABCI_CONFIG_PORT": "26658",
            }

    @staticmethod
    def check_docker_compose_build(
        build_dir: Path,
    ) -> None:
        """Check docker compose build directory."""
        build_tree = list(map(lambda x: x.name, build_dir.iterdir()))
        assert any(
            [
                child in build_tree
                for child in [
                    "persistent_storage",
                    "nodes",
                    DockerComposeGenerator.output_name,
                ]
            ]
        )


class TestLocalhostBuilds(BaseDeployBuildTest):
    """Test localhost builds."""

    def setup(self) -> None:
        """Setup test for localhost deployment."""
        super().setup()
        shutil.copy(
            ROOT_DIR
            / PACKAGES
            / VALORY
            / "agents"
            / "register_reset"
            / DEFAULT_AEA_CONFIG_FILE,
            self.t / "register_reset",
        )
        aea_cli_config = DEFAULT_CLI_CONFIG
        aea_cli_config["registry_config"]["settings"][REGISTRY_LOCAL][
            "default_packages_path"
        ] = (ROOT_DIR / PACKAGES).as_posix()
        Path(CLI_CONFIG_PATH).write_text(yaml.dump(aea_cli_config))

        with open(self.t / "register_reset" / DEFAULT_AEA_CONFIG_FILE, "r") as fp:
            agent_config = next(yaml.safe_load_all(fp))
        agent_config["private_key_paths"]["ethereum"] = ETHEREUM_PRIVATE_KEY_FILE
        with open(self.t / "register_reset" / DEFAULT_AEA_CONFIG_FILE, "w") as fp:
            yaml.dump(agent_config, fp)

        # add all the components
        for component_type in (CONNECTION, CONTRACT, SKILL, PROTOCOL):
            for component_name in agent_config[component_type + "s"]:
                _run_aea_cmd(
                    [
                        "--skip-consistency-check",
                        "add",
                        component_type,
                        component_name,
                        "--mixed",
                    ],
                    cwd=self.t / "register_reset",
                    ignore_error="already exists",
                )

        # prepare ethereum private key
        with open(self.t / "register_reset" / ETHEREUM_PRIVATE_KEY_FILE, "w") as fp:
            fp.write(  # mock private key
                "0x0000000000000000000000000000000000000000000000000000000000000001"
            )

    def test_localhost_build(
        self,
    ) -> None:
        """Test that the build command works."""

        build_dir = self.t / "register_reset" / DEFAULT_BUILD_FOLDER
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (str(self.keys_file), "--localhost", "--mkdir", "data")
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()
        assert (build_dir / "data").exists()

        self.check_localhost_build(build_dir=build_dir)
        self.load_and_check_localhost_build(path=build_dir)


class TestDockerComposeBuilds(BaseDeployBuildTest):
    """Test docker-compose build."""

    def test_docker_compose_build(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        assert not (build_dir / "nodes").exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )
        self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

    def test_docker_compose_build_with_testnet(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                    "--local-tm-setup",
                )
            )

        assert result.exit_code == 0, result.output
        assert (build_dir / "nodes").exists(), result.stderr

    def test_docker_compose_build_log_level(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                    "--log-level",
                    DEBUG,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            f"LOG_LEVEL={DEBUG}"
            in docker_compose["services"][self.spec.get_abci_container_name(0)][
                "environment"
            ]
        )
        assert (
            f"LOG_LEVEL={DEBUG}"
            in docker_compose["services"][self.spec.get_tm_container_name(0)][
                "environment"
            ]
        )

    def test_docker_compose_build_dev(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--dev",
                    "--local",
                    "--packages-dir",
                    str(ROOT_DIR),
                    "--open-aea-dir",
                    str(ROOT_DIR),
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            f"{ROOT_DIR}:/home/ubuntu/packages:rw"
            in docker_compose["services"][self.spec.get_abci_container_name(index=0)][
                "volumes"
            ]
        )
        assert (
            f"{ROOT_DIR}:/open-aea"
            in docker_compose["services"][self.spec.get_abci_container_name(index=0)][
                "volumes"
            ]
        )
        assert (
            f"{ROOT_DIR}:/open-autonomy"
            in docker_compose["services"][self.spec.get_abci_container_name(index=0)][
                "volumes"
            ]
        )

    def test_docker_compose_password(
        self,
    ) -> None:
        """Run tests."""
        keys_file = Path(ETHEREUM_ENCRYPTED_KEYS)
        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())

        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(keys_file),
                    "--o",
                    str(build_dir),
                    "--password",
                    ETHEREUM_ENCRYPTION_PASSWORD,
                    "--local",
                )
            )

        assert result.exit_code == 0, result.stderr
        assert (
            "WARNING: `--password` flag has been deprecated, use `OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD` to export the password value"
            in result.stdout
        )
        assert build_dir.exists()
        assert result.exit_code == 0, result.output

        docker_compose_file = build_dir / DockerComposeGenerator.output_name
        with open(docker_compose_file, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        def _file_check(n: int) -> bool:
            return (
                build_dir
                / DEPLOYMENT_KEY_DIRECTORY
                / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=n)
            ).exists()

        agents = int(len(docker_compose["services"]) / 2)
        assert all(_file_check(i) for i in range(agents))
        for x in range(agents):
            env = dict(
                [
                    f.split("=")
                    for f in docker_compose["services"][
                        self.spec.get_abci_container_name(x)
                    ]["environment"]
                ]
            )
            assert "AEA_PASSWORD" in env.keys()
            assert env["AEA_PASSWORD"] == "$OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD"

    def test_include_acn_and_hardhat_nodes(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--dev",
                    "--local",
                    "--packages-dir",
                    str(ROOT_DIR),
                    "--open-aea-dir",
                    str(ROOT_DIR),
                    "--use-hardhat",
                    "--use-acn",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name,
        )
        assert "acn" in docker_compose["services"]
        assert "hardhat" in docker_compose["services"]

    def test_build_dev_failures(
        self,
    ) -> None:
        """Run tests."""

        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())),
                    "--dev",
                    "--local",
                )
            )

        assert result.exit_code == 1, result.output
        assert "Please provide proper value for --packages-dir" in result.output

    def test_non_existent_keys_file_raises(self) -> None:
        """Test non-existent keys file"""

        keys_file = "non_existent_keys.json"
        result = self.run_cli((keys_file,))
        expected = f"No such file or directory: {Path.cwd() / keys_file}. Please provide valid path for keys file."
        assert result.exit_code == 1
        assert expected in result.stdout

    def test_remove_build_dir_on_exception(self) -> None:
        """Test non-existent keys file"""

        target = "autonomy.cli.deploy.build_deployment"
        side_effect = FileNotFoundError("Mocked file not found")
        with mock.patch(target, side_effect=side_effect) as m:
            with mock.patch("shutil.rmtree") as rmtree_mock:
                result = self.run_cli((str(self.keys_file),))
                m.assert_called_once()
                rmtree_mock.assert_called_once()
                assert str(side_effect) in result.stdout

    def test_docker_compose_build_image_author_flag_default(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                    "--log-level",
                    DEBUG,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            docker_compose["services"][self.spec.get_abci_container_name(0)][
                "image"
            ].split("/")[0]
            == get_default_author_from_cli_config()
            or DEFAULT_DOCKER_IMAGE_AUTHOR
        )

    def test_docker_compose_build_image_author_flag_custom(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        image_author = "some_author"
        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", return_value="test"
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                    "--log-level",
                    DEBUG,
                    "--image-author",
                    image_author,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            docker_compose["services"][self.spec.get_abci_container_name(0)][
                "image"
            ].split("/")[0]
            == image_author
        )

    def test_docker_compose_build_multiple(
        self,
    ) -> None:
        """Run tests."""

        mock_build_hash_id_cycle = cycle(["test", "test2"])

        def mock_build_hash_id() -> str:
            """Mock build hash id."""
            return next(mock_build_hash_id_cycle)

        with mock.patch("os.chown"), OS_ENV_PATCH, mock.patch(
            "autonomy.cli.deploy.build_hash_id", new=mock_build_hash_id
        ):
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--number-of-services",
                    "2",
                )
            )

        assert result.exit_code == 0, result.output

        builds_count = 0
        for build_dir in Path.cwd().glob("abci_build_*"):
            if not (build_dir / DOCKER_COMPOSE_YAML).exists():
                continue

            builds_count += 1

            self.check_docker_compose_build(
                build_dir=build_dir,
            )
            self.load_and_check_docker_compose_file(
                path=build_dir / DockerComposeGenerator.output_name,
                service_hash_id=mock_build_hash_id(),
            )

        assert builds_count == 2


class TestKubernetesBuild(BaseDeployBuildTest):
    """Test kubernetes builds."""

    def test_kubernetes_build(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--local",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)

    def test_kubernetes_build_log_level(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--local",
                    "--log-level",
                    DEBUG,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)

        build_config = self.load_kubernetes_config(build_dir)
        assert {"name": "LOG_LEVEL", "value": DEBUG} in build_config[1]["spec"][
            "template"
        ]["spec"]["containers"][0]["env"]

    def test_kubernetes_build_dev(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--dev",
                    "--local",
                    "--packages-dir",
                    str(ROOT_DIR),
                    "--open-aea-dir",
                    str(ROOT_DIR),
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(
            build_dir=build_dir,
        )

    def test_kubernetes_build_password(
        self,
    ) -> None:
        """Run tests."""
        keys_file = Path(ETHEREUM_ENCRYPTED_KEYS)
        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())

        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--password",
                    ETHEREUM_ENCRYPTION_PASSWORD,
                    "--local",
                )
            )

        assert result.exit_code == 0, result.output
        assert (
            "WARNING: `--password` flag has been deprecated, use `OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD` to export the password value"
            in result.stdout
        )
        assert build_dir.exists()

        kubernetes_config = self.load_kubernetes_config(
            path=build_dir,
        )
        for resource in kubernetes_config:
            try:
                container = resource["spec"]["template"]["spec"]["containers"]
                agent_vars = {f["name"]: f["value"] for f in container[1]["env"]}
            except (KeyError, IndexError):
                continue

            assert agent_vars["AEA_PASSWORD"] == ""

        assert all(
            (
                build_dir
                / DEPLOYMENT_KEY_DIRECTORY
                / KUBERNETES_AGENT_KEY_NAME.format(agent_n=i)
            ).exists()
            for i in range(4)
        )

    def test_kubernetes_build_with_testnet(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())

        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--local",
                    "--kubernetes",
                    "--local-tm-setup",
                )
            )

        assert result.exit_code == 0, result.output

        kubernetes_config = self.load_kubernetes_config(
            path=build_dir,
        )

        assert any(
            [
                resource["metadata"]["name"] == "config-nodes"
                for resource in kubernetes_config
            ]
        )

    def test_kubernetes_build_image_author_default(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--local",
                    "--log-level",
                    DEBUG,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)

        build_config = self.load_kubernetes_config(build_dir)
        assert (
            f"'image': '{get_default_author_from_cli_config() or  DEFAULT_DOCKER_IMAGE_AUTHOR}/oar-"
            in str(build_config)
        )

    def test_kubernetes_build_image_author_custom(
        self,
    ) -> None:
        """Run tests."""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        image_author = "some_image_author"
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--local",
                    "--log-level",
                    DEBUG,
                    "--image-author",
                    image_author,
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)

        build_config = self.load_kubernetes_config(build_dir)
        assert f"'image': '{image_author}/oar-" in str(build_config)
        assert f"'image': '{DEFAULT_DOCKER_IMAGE_AUTHOR}/oar-" not in str(build_config)


class TestExposePorts(BaseDeployBuildTest):
    """Test expose ports from service config."""

    def setup(self) -> None:
        """Setup test."""
        super().setup()

        service_data = get_dummy_service_config(file_number=1)
        service_data[0]["deployment"] = {
            "agent": {"ports": {0: {8080: 8081}}},
            "tendermint": {"ports": {0: {26656: 26666}}},
        }

        with open("./service.yaml", "w+") as fp:
            yaml.dump_all(service_data, fp)

        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset",
                self.keys_file,
            )

    def test_expose_agent_ports_docker_compose(self) -> None:
        """Test expose agent ports"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )
        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert docker_compose["services"][self.spec.get_abci_container_name(0)][
            "ports"
        ] == ["8080:8081"]
        assert docker_compose["services"][self.spec.get_tm_container_name(0)][
            "ports"
        ] == ["26656:26666"]

    def test_expose_agent_ports_kubernetes(self) -> None:
        """Test expose agent ports"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)

        build_config = self.load_kubernetes_config(build_dir)

        assert "ports" in build_config[1]["spec"]["template"]["spec"]["containers"][1]
        assert build_config[1]["spec"]["template"]["spec"]["containers"][1][
            "ports"
        ] == [{"containerPort": 8080}]


class TestExtraVolumes(BaseDeployBuildTest):
    """Test expose ports from service config."""

    volume: Path

    def setup(self) -> None:
        """Setup test."""
        super().setup()

        self.volume = self.t / "extra_volume"
        service_data = get_dummy_service_config(file_number=1)
        service_data[0]["deployment"] = {
            "agent": {"volumes": {f"{self.volume}": "/data"}}
        }

        with open("./service.yaml", "w+") as fp:
            yaml.dump_all(service_data, fp)

        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset",
                self.keys_file,
            )

    def test_expose_agent_ports_docker_compose(self) -> None:
        """Test expose agent ports"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )
        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert self.volume.exists()
        assert (
            f"{self.volume}:/data:Z"
            in docker_compose["services"][self.spec.get_abci_container_name(0)][
                "volumes"
            ]
        )

    def test_expose_agent_ports_kubernetes(self) -> None:
        """Test expose agent ports"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)
        build_config = self.load_kubernetes_config(build_dir)
        assert self.volume.exists()
        assert (
            build_config[1]["spec"]["template"]["spec"]["volumes"][-1]["name"]
            == "extra-volume"
        )
        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1][
                "volumeMounts"
            ][-1]["mountPath"]
            == "/data"
        )
        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1][
                "volumeMounts"
            ][-1]["name"]
            == "extra-volume"
        )
        assert build_config[-1]["metadata"]["name"] == "extra-volume"


class TestLoadEnvVars(BaseDeployBuildTest):
    """Test expose ports from service config."""

    env_var = "SERVICE_HELLO_WORLD_HELLO_WORLD_STRING"
    env_var_path = "SKILL_DUMMY_SKILL_MODELS_PARAMS_ARGS_HELLO_WORLD_MESSAGE"
    env_var_value = "ENV_VAR_VALUE"

    def setup(self) -> None:
        """Setup test."""
        super().setup()
        service_data = get_dummy_service_config(file_number=4)
        with open("./service.yaml", "w+") as fp:
            yaml.dump_all(service_data, fp)
        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset",
                self.keys_file,
            )

    def _run_test(self) -> None:
        """Run test."""
        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )

        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            f"{self.env_var_path}={self.env_var_value}"
            in docker_compose["services"][self.spec.get_abci_container_name(0)][
                "environment"
            ]
        )

    def test_load_dot_env(self) -> None:
        """Test load `.env` file"""

        (self.t / "register_reset" / DEFAULT_ENV_DOTFILE).write_text(
            f"{self.env_var}={self.env_var_value}"
        )
        self._run_test()

    def test_load_json(self) -> None:
        """Test load `.json` file"""

        env_var_value = "ENV_VAR_VALUE"
        env_file = self.t / "register_reset" / "env.json"
        env_file.write_text(json.dumps({self.env_var: env_var_value}))
        self.cli_options = ("deploy", "--env-file", str(env_file.resolve()), "build")

        self._run_test()


class TestResourceSpecification(BaseDeployBuildTest):
    """Test expose ports from service config."""

    def setup(self) -> None:
        """Setup test."""
        super().setup()
        service_data = get_dummy_service_config(file_number=4)
        with open("./service.yaml", "w+") as fp:
            yaml.dump_all(service_data, fp)
        with OS_ENV_PATCH:
            self.spec = ServiceBuilder.from_dir(
                self.t / "register_reset",
                self.keys_file,
            )

    def test_default_resources_docker_compose(self) -> None:
        """Test custom resources"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli((str(self.keys_file), "--o", str(build_dir)))

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )
        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            docker_compose["services"]["dummyservice_abci_0"]["mem_reservation"]
            == f"{DEFAULT_AGENT_MEMORY_REQUEST}M"
        )
        assert (
            docker_compose["services"]["dummyservice_abci_0"]["mem_limit"]
            == f"{DEFAULT_AGENT_MEMORY_LIMIT}M"
        )
        assert (
            docker_compose["services"]["dummyservice_abci_0"]["cpus"]
            == DEFAULT_AGENT_CPU_LIMIT
        )

    def test_default_resources_kuberntes(self) -> None:
        """Test custom resources"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)
        build_config = self.load_kubernetes_config(build_dir)

        assert build_config[1]["spec"]["template"]["spec"]["containers"][1][
            "resources"
        ]["requests"]["cpu"] == str(DEFAULT_AGENT_CPU_REQUEST)

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "requests"
            ]["memory"]
            == f"{DEFAULT_AGENT_MEMORY_REQUEST}Mi"
        )

        assert build_config[1]["spec"]["template"]["spec"]["containers"][1][
            "resources"
        ]["limits"]["cpu"] == str(DEFAULT_AGENT_CPU_LIMIT)

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "limits"
            ]["memory"]
            == f"{DEFAULT_AGENT_MEMORY_LIMIT}Mi"
        )

    def test_custom_resources_docker_compose(self) -> None:
        """Test custom resources"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--agent-cpu-limit",
                    "2.0",
                    "--agent-memory-limit",
                    "4096",
                    "--agent-memory-request",
                    "2048",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_docker_compose_build(
            build_dir=build_dir,
        )
        docker_compose = self.load_and_check_docker_compose_file(
            path=build_dir / DockerComposeGenerator.output_name
        )

        assert (
            docker_compose["services"]["dummyservice_abci_0"]["mem_reservation"]
            == "2048M"
        )
        assert docker_compose["services"]["dummyservice_abci_0"]["mem_limit"] == "4096M"
        assert docker_compose["services"]["dummyservice_abci_0"]["cpus"] == 2.0

    def test_custom_resources_kuberntes(self) -> None:
        """Test custom resources"""

        build_dir = self.t / DEFAULT_BUILD_FOLDER.format(build_hash_id())
        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.run_cli(
                (
                    str(self.keys_file),
                    "--o",
                    str(build_dir),
                    "--kubernetes",
                    "--agent-cpu-request",
                    "1.0",
                    "--agent-memory-request",
                    "2048",
                    "--agent-cpu-limit",
                    "2.0",
                    "--agent-memory-limit",
                    "4096",
                )
            )

        assert result.exit_code == 0, result.output
        assert build_dir.exists()

        self.check_kubernetes_build(build_dir=build_dir)
        build_config = self.load_kubernetes_config(build_dir)

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "requests"
            ]["cpu"]
            == "1.0"
        )

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "requests"
            ]["memory"]
            == "2048Mi"
        )

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "limits"
            ]["cpu"]
            == "2.0"
        )

        assert (
            build_config[1]["spec"]["template"]["spec"]["containers"][1]["resources"][
                "limits"
            ]["memory"]
            == "4096Mi"
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools for aggregating benchmark results."""

import json
import statistics
from pathlib import Path
from typing import Callable, Dict, List, Tuple, cast

from autonomy.analyse.benchmark.html import (
    BLOCK_TEMPLATE,
    HTML_TEMPLATE,
    TABLE_TEMPLATE,
    TD_TEMPLATE,
    TH_TEMPLATE,
    TROW_TEMPLATE,
)


STATISTICS = {
    "Mean": statistics.mean,
    "Maximum": max,
    "Minimum": min,
}


class BlockTypes:  # pylint: disable=too-few-public-methods
    """Block types."""

    ALL = "all"
    LOCAL = "local"
    CONSENSUS = "consensus"
    TOTAL = "total"

    types = (LOCAL, CONSENSUS, TOTAL)


def read_benchmark_data(
    path: Path,
    block_type: str = BlockTypes.ALL,
    period: int = -1,
) -> Dict[str, Dict[str, List[Dict]]]:
    """Returns logs."""
    benchmark_data: Dict[str, Dict[str, List[Dict]]] = {}
    for agent_dir in path.iterdir():
        benchmark_data[agent_dir.name] = {}
        benchmark_data_files = list(agent_dir.glob("*.json"))
        periods = sorted(
            tuple(map(lambda x: x.name, benchmark_data_files))
            if period == -1
            else (f"{period}.json",)
        )
        for _period in periods:
            period_name, _ = _period.split(".")
            period_data = json.loads((agent_dir / _period).read_text())
            if block_type == BlockTypes.ALL:
                benchmark_data[agent_dir.name][period_name] = period_data
                continue

            benchmark_data[agent_dir.name][period_name] = [
                {
                    "behaviour": behaviour["behaviour"],
                    "data": {block_type: behaviour["data"][block_type]},
                }
                for behaviour in period_data
            ]

    return benchmark_data


def add_statistic(
    name: str,
    aggregator: Callable,
    behaviours: List[str],
    behaviour_history: Dict[str, List[float]],
) -> str:
    """Add a stastic column."""
    rows = TD_TEMPLATE.format(name)
    for behaviour in behaviours:
        rows += TD_TEMPLATE.format(aggregator(behaviour_history[behaviour]))
    return TROW_TEMPLATE.format(rows)


def add_statistics(
    behaviours: List[str],
    behaviour_history: Dict[str, List[float]],
) -> str:
    """Add statistics."""
    tbody = TROW_TEMPLATE.format(
        f"""<td style="text-align: center;" colspan="{len(behaviours)}">Statistics</td>"""
    )
    for name, aggregator in STATISTICS.items():
        tbody += add_statistic(
            name=name,
            aggregator=cast(Callable, aggregator),
            behaviour_history=behaviour_history,
            behaviours=behaviours,
        )
    return tbody


def create_table_data(
    data: Dict[str, List[Dict]],
    blocks: Tuple[str, ...],
) -> Tuple[List[str], List[str], Dict[str, Dict]]:
    """Create table data."""
    periods: List[str] = []
    header_columns: List[str] = ["Period"]
    tables_data: Dict[str, Dict] = {}
    for block in blocks:
        if block not in tables_data:
            tables_data[block] = {}
        for period, behaviours in data.items():
            for behaviour in behaviours:
                if period not in tables_data[block]:
                    tables_data[block][period] = {}
                tables_data[block][period][behaviour["behaviour"]] = behaviour["data"][
                    block
                ]
                if behaviour["behaviour"] not in header_columns:
                    header_columns.append(behaviour["behaviour"])
            if period not in periods:
                periods.append(period)
    return periods, header_columns, tables_data


def create_agent_table(
    agent: str, data: Dict[str, List[Dict]], blocks: Tuple[str, ...]
) -> str:
    """Create agent table."""
    tables = f"""<div style="width: 100%; text-align: center"> Benchmark data for agent {agent}</div>\n"""
    periods, header_columns, tables_data = create_table_data(
        data=data,
        blocks=blocks,
    )
    behaviour_history: Dict[str, List[float]] = {
        behaviour: [] for behaviour in header_columns[1:]
    }
    for block, block_data in tables_data.items():
        thead = "".join(map(TH_TEMPLATE.format, header_columns))
        tbody = ""
        for period in sorted(periods):
            rows = TD_TEMPLATE.format(period)
            for behaviour_name in header_columns[1:]:
                behaviour_history[behaviour_name].append(
                    block_data[period][behaviour_name]
                )
                rows += TD_TEMPLATE.format(block_data[period][behaviour_name])
            tbody += TROW_TEMPLATE.format(rows)
        tbody += add_statistics(
            behaviours=header_columns[1:], behaviour_history=behaviour_history
        )
        tables += BLOCK_TEMPLATE.format(
            table=TABLE_TEMPLATE.format(
                colspan=len(header_columns),
                block_type=block,
                thead=thead,
                tbody=tbody,
            ),
        )
    return tables


def aggregate(path: Path, block_type: str, period: int, output: Path) -> None:
    """Benchmark Aggregator."""
    path = Path(path)
    benchmark_data = read_benchmark_data(
        path,
        block_type=block_type,
        period=period,
    )
    tables = [
        create_agent_table(
            agent=agent,
            data=data,
            blocks=(
                BlockTypes.types if block_type == BlockTypes.ALL else (block_type,)
            ),
        )
        for agent, data in benchmark_data.items()
    ]

    output.write_text(HTML_TEMPLATE.format(tables="".join(tables)), encoding="utf-8")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads of the DummyAbciApp."""

from dataclasses import dataclass

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


@dataclass(frozen=True)
class DummyStartingPayload(BaseTxPayload):
    """Represent a transaction payload for the DummyStartingRound."""

    content: str


@dataclass(frozen=True)
class DummyRandomnessPayload(BaseTxPayload):
    """Represent a transaction payload for the DummyRandomnessRound."""

    round_id: int
    randomness: str


@dataclass(frozen=True)
class DummyKeeperSelectionPayload(BaseTxPayload):
    """Represent a transaction payload for the DummyKeeperSelectionRound."""

    keepers: str


@dataclass(frozen=True)
class DummyFinalPayload(BaseTxPayload):
    """Represent a transaction payload for the DummyFinalRound."""

    content: bool



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the slashing rounds."""

import json
from enum import Enum
from typing import Dict, List, Optional, Set, Tuple, cast

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AppState,
    BaseSynchronizedData,
    BaseTxPayload,
    CollectSameUntilThresholdRound,
    CollectionRound,
    DeserializedCollection,
    SELECTION_KEY_ATTRIBUTE,
    TransactionNotValidError,
    get_name,
)
from packages.valory.skills.slashing_abci.payloads import (
    SlashingTxPayload,
    StatusResetPayload,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSyncedData,
)


class Event(Enum):
    """Defines the round events."""

    SLASH_START = "slash_start"
    SLASH_END = "slash_end"
    NO_MAJORITY = "no_majority"
    ROUND_TIMEOUT = "round_timeout"
    NONE = "none"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def slashing_bg_init(self) -> bool:
        """Get if the slashing background round has been initialized."""
        return bool(self.db.get("slashing_bg_init", False))

    @property
    def slashing_in_flight(self) -> bool:
        """Get if there is a slashing operation in progress."""
        return bool(self.db.get("slashing_in_flight", False))

    @property
    def slashing_majority_reached(self) -> bool:
        """Get if the slashing majority has been reached."""
        return bool(self.db.get("slashing_majority_reached", False))

    @property
    def operators_mapping(self) -> Optional[Dict[str, List[str]]]:
        """Get a mapping of the operators mapped to their agent instances."""
        mapping = self.db.get("operators_mapping", None)

        if mapping is None:
            return None

        return json.loads(mapping)

    @property
    def slash_timestamps(self) -> Dict[str, int]:
        """Get the timestamp in which each agent instance was slashed at."""
        timestamps = str(self.db.get("slash_timestamps", ""))

        if timestamps == "":
            return {}

        return json.loads(timestamps)

    @property
    def participant_to_offence_reset(
        self,
    ) -> DeserializedCollection:  # pragma: no cover
        """The participants mapped to the status reset payloads."""
        serialized = self.db.get_strict("participant_to_randomness")
        deserialized = CollectionRound.deserialize_collection(serialized)
        return cast(DeserializedCollection, deserialized)


class SlashingCheckRound(CollectSameUntilThresholdRound):
    """Defines the slashing check background round, which runs concurrently with other rounds to send the slash tx."""

    payload_class = SlashingTxPayload
    synchronized_data_class = SynchronizedData
    selection_key = (
        get_name(TxSettlementSyncedData.most_voted_tx_hash),
        get_name(SynchronizedData.slashing_in_flight),
        get_name(SynchronizedData.slashing_majority_reached),
    )
    extended_requirements: Tuple[str, ...] = (SELECTION_KEY_ATTRIBUTE,)

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""
        # for background round payloads, we don't need to check the round_count, as round_count is irrelevant for the
        # background since it's running concurrently in the background.
        sender = payload.sender
        if sender not in self.accepting_payloads_from:
            raise ABCIAppInternalError(
                f"{sender} not in list of participants: {sorted(self.accepting_payloads_from)}"
            )

        if sender in self.collection:
            raise ABCIAppInternalError(
                f"sender {sender} has already sent value for round: {self.round_id}"
            )

        self.collection[sender] = payload

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check Payload"""
        # NOTE: the TransactionNotValidError is intercepted in ABCIRoundHandler.deliver_tx
        #  which means it will be logged instead of raised
        # for background round payloads, we don't need to check the round_count, as round_count is irrelevant for the
        # background since it's running concurrently in the background.
        sender_in_participant_set = payload.sender in self.accepting_payloads_from
        if not sender_in_participant_set:
            raise TransactionNotValidError(
                f"{payload.sender} not in list of participants: {sorted(self.accepting_payloads_from)}"
            )

        if payload.sender in self.collection:
            raise TransactionNotValidError(
                f"sender {payload.sender} has already sent value for round: {self.round_id}"
            )

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        synced_data = SynchronizedData(db=self.synchronized_data.db)
        if not synced_data.slashing_bg_init:
            # we set the cross-period values because the period will most probably be increased
            # before this background round ever returns a `SLASH_START` event.
            # if this happens, and we do not set them here, then the db creation for the new period will fail
            # because the values are in the cross-period persisted keys.
            self.synchronized_data.update(
                synchronized_data_class=self.synchronized_data_class,
                slashing_bg_init=True,
                operators_mapping=None,
                slash_timestamps="",
                slashing_in_flight=synced_data.slashing_in_flight,
                slashing_majority_reached=synced_data.slashing_majority_reached,
            )

        if not self.threshold_reached or synced_data.slashing_in_flight:
            return None

        majority_data = dict(zip(self.selection_key, self.most_voted_payload_values))
        state = synced_data.update(
            synchronized_data_class=self.synchronized_data_class,
            **majority_data,
        )
        # we need to reset the collection before returning the slash start event,
        # because this round is continuously and concurrently run in the background
        self.collection = {}
        return state, Event.SLASH_START


class StatusResetRound(CollectSameUntilThresholdRound):
    """Defines the slashing check background round, which runs after a slash tx has been verified."""

    payload_class = StatusResetPayload
    synchronized_data_class = SynchronizedData
    collection_key = get_name(SynchronizedData.participant_to_offence_reset)
    selection_key = (
        get_name(SynchronizedData.operators_mapping),
        get_name(SynchronizedData.slash_timestamps),
        get_name(SynchronizedData.slashing_in_flight),
        get_name(SynchronizedData.slashing_majority_reached),
    )
    done_event = Event.SLASH_END
    no_majority_event = Event.NO_MAJORITY
    none_event = Event.NONE


class PostSlashingTxAbciApp(AbciApp[Event]):
    """PostSlashingTxAbciApp

    Initial round: StatusResetRound

    Initial states: {StatusResetRound}

    Transition states:
        0. StatusResetRound
            - slash end: 0.
            - no majority: 0.
            - round timeout: 0.
            - none: 0.

    Final states: {}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls = StatusResetRound
    initial_states = {StatusResetRound}
    transition_function = {
        StatusResetRound: {
            # the following is not needed, it is added to satisfy the round check.
            # the `SLASH_END` event is the end event of the slashing background app,
            # which signals the app to return to the main transition function
            # for more information, see `BackgroundApp` and `AbciApp` implementation
            Event.SLASH_END: StatusResetRound,
            Event.NO_MAJORITY: StatusResetRound,
            Event.ROUND_TIMEOUT: StatusResetRound,
            # none event cannot occur
            Event.NONE: StatusResetRound,
        },
    }
    cross_period_persisted_keys = frozenset(
        {
            get_name(SynchronizedData.operators_mapping),
            get_name(SynchronizedData.slash_timestamps),
            get_name(SynchronizedData.slashing_in_flight),
            get_name(SynchronizedData.slashing_majority_reached),
        }
    )
    event_to_timeout: Dict[Event, float] = {Event.ROUND_TIMEOUT: 30.0}
    db_pre_conditions: Dict[AppState, Set[str]] = {StatusResetRound: set()}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/crypto/keys.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1ctendermint/crypto/keys.proto\x12\x11tendermint.crypto\x1a\x14gogoproto/gogo.proto"D\n\tPublicKey\x12\x11\n\x07\x65\x64\x32\x35\x35\x31\x39\x18\x01 \x01(\x0cH\x00\x12\x13\n\tsecp256k1\x18\x02 \x01(\x0cH\x00:\x08\xe8\xa0\x1f\x01\xe8\xa1\x1f\x01\x42\x05\n\x03sumB:Z8github.com/tendermint/tendermint/proto/tendermint/cryptob\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "tendermint.crypto.keys_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z8github.com/tendermint/tendermint/proto/tendermint/crypto"
    )
    _PUBLICKEY._options = None
    _PUBLICKEY._serialized_options = b"\350\240\037\001\350\241\037\001"
    _globals["_PUBLICKEY"]._serialized_start = 73
    _globals["_PUBLICKEY"]._serialized_end = 141
# @@protoc_insertion_point(module_scope)
